import synapse.common as s_common

import synapse.tests.utils as s_t_utils

class DnsModelTest(s_t_utils.SynTest):

    async def test_model_dns_name_type(self):
        async with self.getTestCore() as core:
            typ = core.model.type('inet:dns:name')
            # ipv4 - good and newp
            norm, info = typ.norm('4.3.2.1.in-addr.ARPA')
            self.eq(norm, '4.3.2.1.in-addr.arpa')
            self.eq(info.get('subs'), {'ipv4': 0x01020304})
            norm, info = typ.norm('newp.in-addr.ARPA')
            self.eq(norm, 'newp.in-addr.arpa')
            self.eq(info.get('subs'), {})

            # Ipv6 - good, newp, and ipv4 included
            ipv6 = 'b.a.9.8.7.6.5.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.8.b.d.0.1.0.0.2.ip6.ARPA'
            norm, info = typ.norm(ipv6)
            self.eq(norm, 'b.a.9.8.7.6.5.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.8.b.d.0.1.0.0.2.ip6.arpa')
            self.eq(info.get('subs'), {'ipv6': '2001:db8::567:89ab'})

            ipv6 = 'newp.2.ip6.arpa'
            norm, info = typ.norm(ipv6)
            self.eq(norm, 'newp.2.ip6.arpa')
            self.eq(info.get('subs'), {})

            ipv6 = '4.0.3.0.2.0.1.0.f.f.f.f.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.ip6.arpa'
            norm, info = typ.norm(ipv6)
            self.eq(norm, ipv6)
            self.eq(info.get('subs'), {'ipv6': '::ffff:1.2.3.4', 'ipv4': 0x01020304})

            # fqdn and a invalid fqdn
            norm, info = typ.norm('test.vertex.link')
            self.eq(norm, 'test.vertex.link')
            self.eq(info.get('subs'), {'fqdn': 'test.vertex.link'})

            norm, info = typ.norm('1.2.3.4')
            self.eq(norm, '1.2.3.4')
            self.eq(info.get('subs'), {'ipv4': 0x01020304})

            norm, info = typ.norm('134744072')  # 8.8.8.8 in integer form
            self.eq(norm, '134744072')
            self.eq(info.get('subs'), {})

            norm, info = typ.norm('::FFFF:1.2.3.4')
            self.eq(norm, '::ffff:1.2.3.4')
            self.eq(info.get('subs'), {'ipv6': '::ffff:1.2.3.4', 'ipv4': 0x01020304})

            norm, info = typ.norm('::1')
            self.eq(norm, '::1')
            self.eq(info.get('subs'), {'ipv6': '::1'})

    async def test_model_dns_request(self):

        async with self.getTestCore() as core:

            async with await core.snap() as snap:

                props = {
                    'time': '2018',
                    'query': ('1.2.3.4', 'vertex.link', 255),
                    'server': 'udp://5.6.7.8:53',
                    'reply:code': 0,
                }
                expected_nodes = (
                    ('inet:server', 'udp://5.6.7.8:53'),
                    ('inet:fqdn', 'vertex.link'),
                )
                node = await snap.addNode('inet:dns:request', '*', props)
                req_ndef = node.ndef
                self.eq(node.get('time'), 1514764800000)
                self.eq(node.get('reply:code'), 0)
                self.eq(node.get('server'), 'udp://5.6.7.8:53')
                self.eq(node.get('query'), ('tcp://1.2.3.4', 'vertex.link', 255))
                self.eq(node.get('query:name'), 'vertex.link')
                self.eq(node.get('query:name:fqdn'), 'vertex.link')
                self.eq(node.get('query:type'), 255)
                self.none(node.get('query:client'))
                await self.checkNodes(core, expected_nodes)

                # Ensure some remaining inet:dns:query:name:* props are broken out
                node = await snap.addNode('inet:dns:request', '*', {'query:name': '4.3.2.1.in-addr.arpa'})
                self.none(node.get('query:name:fqdn'))
                self.eq(node.get('query:name:ipv4'), 0x01020304)
                self.eq(node.get('query:name'), '4.3.2.1.in-addr.arpa')

                # A bit of a bunk example but sometimes people query for raw ipv4/ipv6 addresses
                # and we'll try to extract them if possible :)
                node = await snap.addNode('inet:dns:request', '*', {'query:name': '::ffff:1.2.3.4'})
                self.none(node.get('query:name:fqdn'))
                self.eq(node.get('query:name'), '::ffff:1.2.3.4')
                self.eq(node.get('query:name:ipv4'), 0x01020304)
                self.eq(node.get('query:name:ipv6'), '::ffff:1.2.3.4')

                # Ensure that lift via prefix for inet:dns:name type works
                nodes = await snap.nodes('inet:dns:request:query:name^=vertex')
                self.len(1, nodes)

                # Ensure that subs are broken out for inet:dns:query
                node = await snap.getNodeByNdef(('inet:dns:query', ('tcp://1.2.3.4', 'vertex.link', 255)))
                self.eq(node.get('client'), 'tcp://1.2.3.4')
                self.eq(node.get('name'), 'vertex.link')
                self.eq(node.get('name:fqdn'), 'vertex.link')
                self.eq(node.get('type'), 255)

                node = await snap.addNode('inet:dns:query', ('tcp://1.2.3.4', '4.3.2.1.in-addr.arpa', 255))
                self.eq(node.get('name'), '4.3.2.1.in-addr.arpa')
                self.none(node.get('name:fqdn'))
                self.eq(node.get('name:ipv4'), 0x01020304)
                self.none(node.get('name:ipv6'))
                valu = ('tcp://1.2.3.4',
                        '4.0.3.0.2.0.1.0.f.f.f.f.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.ip6.arpa',
                        255)
                node = await snap.addNode('inet:dns:query', valu)
                self.none(node.get('name:fqdn'))
                self.eq(node.get('name:ipv4'), 0x01020304)
                self.eq(node.get('name:ipv6'), '::ffff:1.2.3.4')

                # Try inet:dns:answer now
                props = {
                    'request': req_ndef[1],
                    'a': ('vertex.link', '2.3.4.5'),
                }

                await snap.addNode('inet:dns:answer', '*', props)
                self.nn(await snap.getNodeByNdef(('inet:dns:a', ('vertex.link', 0x02030405))))

                # It is also possible for us to record a request from imperfect data
                # An example of that is dns data from a malware sandbox where the client
                # IP is unknown
                props = {
                    'time': '2018',
                    'exe': f'guid:{"a" * 32}',
                    'query:name': 'notac2.someone.com'
                }
                node = await snap.addNode('inet:dns:request', '*', props)
                self.none(node.get('query'))
                self.eq(node.get('exe'), f'guid:{"a" * 32}')
                self.eq(node.get('query:name'), 'notac2.someone.com')

            nodes = await core.nodes('[inet:dns:request=(test,) :query:name="::ffff:8.7.6.5"]')
            self.len(1, nodes)
            expected_nodes = (
                ('inet:ipv4', 0x08070605),
                ('inet:ipv6', '::ffff:8.7.6.5'),
            )
            await self.checkNodes(core, expected_nodes)

            # DNS queries can be quite complex or awkward since the protocol
            # allows for nearly anything to be asked about. This can lead to
            # pivots with non-normable data.
            q = '[inet:dns:query=(tcp://1.2.3.4, "", 1)]'
            await self.agenlen(1, core.eval(q))
            q = '[inet:dns:query=(tcp://1.2.3.4, "foo*.haha.com", 1)]'
            await self.agenlen(1, core.eval(q))
            q = 'inet:dns:query=(tcp://1.2.3.4, "", 1) :name -> inet:fqdn'
            with self.getLoggerStream('synapse.lib.ast',
                                      'Cannot generate fqdn index bytes for a empty string') as stream:
                await self.agenlen(0, core.eval(q))
                self.true(stream.wait(1))

            q = 'inet:dns:query=(tcp://1.2.3.4, "foo*.haha.com", 1) :name -> inet:fqdn'
            with self.getLoggerStream('synapse.lib.ast',
                                      'Wild card may only appear at the beginning') as stream:
                await self.agenlen(0, core.eval(q))
                self.true(stream.wait(1))

    async def test_forms_dns_simple(self):

        async with self.getTestCore() as core:

            async with await core.snap() as snap:
                # inet:dns:a
                node = await snap.addNode('inet:dns:a', ('hehe.com', '1.2.3.4'))
                self.eq(node.ndef[1], ('hehe.com', 0x01020304))
                self.eq(node.get('fqdn'), 'hehe.com')
                self.eq(node.get('ipv4'), 0x01020304)

                node = await snap.addNode('inet:dns:a', ('www.\u0915\u0949\u092e.com', '1.2.3.4'))
                self.eq(node.ndef[1], ('www.xn--11b4c3d.com', 0x01020304))
                self.eq(node.get('fqdn'), 'www.xn--11b4c3d.com')
                self.eq(node.get('ipv4'), 0x01020304)

                # inet:dns:aaaa
                node = await snap.addNode('inet:dns:aaaa', ('localhost', '::1'))
                self.eq(node.ndef[1], ('localhost', '::1'))
                self.eq(node.get('fqdn'), 'localhost')
                self.eq(node.get('ipv6'), '::1')

                node = await snap.addNode('inet:dns:aaaa', ('hehe.com', '2001:0db8:85a3:0000:0000:8a2e:0370:7334'))
                self.eq(node.ndef[1], ('hehe.com', '2001:db8:85a3::8a2e:370:7334'))
                self.eq(node.get('fqdn'), 'hehe.com')
                self.eq(node.get('ipv6'), '2001:db8:85a3::8a2e:370:7334')

                # inet:dns:rev
                node = await snap.addNode('inet:dns:rev', ('1.2.3.4', 'bebe.com'))
                self.eq(node.ndef[1], (0x01020304, 'bebe.com'))
                self.eq(node.get('ipv4'), 0x01020304)
                self.eq(node.get('fqdn'), 'bebe.com')

                # inet:dns:rev6
                node = await snap.addNode('inet:dns:rev6', ('FF::56', 'bebe.com'))
                self.eq(node.ndef[1], ('ff::56', 'bebe.com'))
                self.eq(node.get('ipv6'), 'ff::56')
                self.eq(node.get('fqdn'), 'bebe.com')

                # inet:dns:ns
                node = await snap.addNode('inet:dns:ns', ('haha.com', 'ns1.haha.com'))
                self.eq(node.ndef[1], ('haha.com', 'ns1.haha.com'))
                self.eq(node.get('zone'), 'haha.com')
                self.eq(node.get('ns'), 'ns1.haha.com')

                # inet:dns:cname
                node = await snap.addNode('inet:dns:cname', ('HAHA.vertex.link', 'vertex.link'))
                self.eq(node.ndef[1], ('haha.vertex.link', 'vertex.link'))
                self.eq(node.get('fqdn'), 'haha.vertex.link')
                self.eq(node.get('cname'), 'vertex.link')

                # inet:dns:mx
                node = await snap.addNode('inet:dns:mx', ('vertex.link', 'mail.vertex.link'))
                self.eq(node.ndef[1], ('vertex.link', 'mail.vertex.link'))
                self.eq(node.get('fqdn'), 'vertex.link')
                self.eq(node.get('mx'), 'mail.vertex.link')

                # inet:dns:soa
                guid = s_common.guid()
                props = {'fqdn': 'haha.vertex.link', 'ns': 'ns1.vertex.link', 'email': 'pennywise@vertex.link'}
                node = await snap.addNode('inet:dns:soa', guid, props)
                self.eq(node.get('fqdn'), 'haha.vertex.link')
                self.eq(node.get('email'), 'pennywise@vertex.link')
                self.eq(node.get('ns'), 'ns1.vertex.link')

                # inet:dns:txt
                node = await snap.addNode('inet:dns:txt', ('clowns.vertex.link', 'we all float down here'))
                self.eq(node.ndef[1], ('clowns.vertex.link', 'we all float down here'))
                self.eq(node.get('fqdn'), 'clowns.vertex.link')
                self.eq(node.get('txt'), 'we all float down here')

    # The inet:dns:answer form has a large number of properties on it,
    async def test_model_inet_dns_answer(self):
        ip0 = 0x01010101
        ip1 = '::2'
        fqdn0 = 'woot.com'
        fqdn1 = 'haha.com'
        email0 = 'pennywise@vertex.ninja'

        async with self.getTestCore() as core:

            async with await core.snap() as snap:
                # a record
                props = {'a': (fqdn0, ip0)}
                node = await snap.addNode('inet:dns:answer', '*', props)
                self.eq(node.get('a'), (fqdn0, ip0))
                # ns record
                props = {'ns': (fqdn0, fqdn1)}
                node = await snap.addNode('inet:dns:answer', '*', props)
                self.eq(node.get('ns'), (fqdn0, fqdn1))
                # rev record
                props = {'rev': (ip0, fqdn0)}
                node = await snap.addNode('inet:dns:answer', '*', props)
                self.eq(node.get('rev'), (ip0, fqdn0))
                # aaaa record
                props = {'aaaa': (fqdn0, ip1)}
                node = await snap.addNode('inet:dns:answer', '*', props)
                self.eq(node.get('aaaa'), (fqdn0, ip1))
                # rev6 record
                props = {'rev6': (ip1, fqdn0)}
                node = await snap.addNode('inet:dns:answer', '*', props)
                self.eq(node.get('rev6'), (ip1, fqdn0))
                # cname record
                props = {'cname': (fqdn0, fqdn1)}
                node = await snap.addNode('inet:dns:answer', '*', props)
                self.eq(node.get('cname'), (fqdn0, fqdn1))
                # mx record
                props = {'mx': (fqdn0, fqdn1)}
                node = await snap.addNode('inet:dns:answer', '*', props)
                self.eq(node.get('mx'), (fqdn0, fqdn1))
                # soa record
                guid = s_common.guid((fqdn0, fqdn1, email0))
                props = {'soa': guid}
                node = await snap.addNode('inet:dns:answer', '*', props)
                self.eq(node.get('soa'), guid)
                # txt record
                props = {'txt': (fqdn0, 'Oh my!')}
                node = await snap.addNode('inet:dns:answer', '*', props)
                self.eq(node.get('txt'), (fqdn0, 'Oh my!'))

    async def test_model_dns_wild(self):

        async with self.getTestCore() as core:

            async with await core.snap() as snap:

                wild = await snap.addNode('inet:dns:wild:a', ('vertex.link', '1.2.3.4'))
                self.eq(wild.ndef, ('inet:dns:wild:a', ('vertex.link', 0x01020304)))
                self.eq(wild.get('ipv4'), 0x01020304)
                self.eq(wild.get('fqdn'), 'vertex.link')

                wild = await snap.addNode('inet:dns:wild:aaaa', ('vertex.link', '2001:db8:85a3::8a2e:370:7334'))
                self.eq(wild.ndef, ('inet:dns:wild:aaaa', ('vertex.link', '2001:db8:85a3::8a2e:370:7334')))
                self.eq(wild.get('ipv6'), '2001:db8:85a3::8a2e:370:7334')
                self.eq(wild.get('fqdn'), 'vertex.link')



import os
import sys
import json
import argparse

import synapse.lib.output as s_output
import synapse.lib.msgpack as s_msgpack

def getArgParser():
    pars = argparse.ArgumentParser(description='Convert files from json lines to msgpack')
    pars.add_argument('--rm', action='store_true', help='Remove json files once the conversion is complete')
    pars.add_argument('paths', nargs='+', help='json files or directories full of json files')
    return pars

def main(argv, outp=None):

    if outp is None:  # pragma: no cover
        outp = s_output.OutPut()

    pars = getArgParser()
    opts = pars.parse_args(argv)

    for path in opts.paths:

        if not path.endswith('.json'):
            outp.printf('skip: %s (not .json extension)' % (path,))
            continue

        if not os.path.isfile(path):
            outp.printf('skip: %s (not a file)' % (path,))
            continue

        base = path[:-5]
        newp = base + '.mpk'

        outp.printf('converting: %s -> .mpk' % (path,))
        with open(path, 'r', encoding='utf8') as fd:
            with open(newp, 'wb') as pk:
                for line in fd:
                    item = json.loads(line)
                    pk.write(s_msgpack.en(item))

        if opts.rm:
            os.unlink(path)

if __name__ == '__main__':  # pragma: no cover
    sys.exit(main(sys.argv[1:]))



import synapse.lib.module as s_module

class TransportModule(s_module.CoreModule):
    def getModelDefs(self):
        modl = {
            'types': (

                ('transport:air:craft', ('guid', {}), {
                    'doc': 'An individual aircraft.'}),

                ('transport:air:tailnum', ('str', {'lower': True, 'strip': True, 'regex': '^[a-z0-9-]{2,}$'}), {
                    'doc': 'An aircraft registration number or military aircraft serial number.',
                    'ex': 'ff023'}),

                ('transport:air:flightnum', ('str', {'lower': True, 'strip': True, 'replace': ((' ', ''),), 'regex': '^[a-z]{2}[0-9]{1,4}$'}), {
                    'doc': 'A commercial flight designator including airline and serial.',
                    'ex': 'ua2437'}),

                ('transport:air:telem', ('guid', {}), {
                    'doc': 'A telemtry sample from an aircraft in transit.'}),

                ('transport:air:flight', ('guid', {}), {
                    'doc': 'An individual instance of a flight.'}),

                ('transport:air:occupant', ('guid', {}), {
                    'doc': 'An occupant of a specific flight.'}),

                ('transport:air:port', ('str', {'lower': True}), {
                    'doc': 'An IATA assigned airport code.'}),

                ('transport:sea:vessel', ('guid', {}), {
                    'doc': 'An individual sea vessel.'}),

                ('transport:sea:mmsi', ('str', {'regex': '[0-9]{9}'}), {
                    'doc': 'A Maritime Mobile Service Identifier'}),

                ('transport:sea:imo', ('str', {'lower': True, 'strip': True, 'replace': ((' ', ''),), 'regex': '^imo[0-9]{7}$'}), {
                    'doc': 'An International Maritime Organization registration number.'}),

                ('transport:sea:telem', ('guid', {}), {
                    'doc': 'A telemetry sample from a vessel in transit.'}),

                # TODO a few more items to plumb eventually
                # ('transport:sea:hin',
                # ('transport:sea:port',
            ),
            'forms': (
                ('transport:air:craft', {}, (
                    ('tailnum', ('transport:air:tailnum', {}), {
                        'doc': 'The aircraft tail number.'}),
                    ('type', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'The type of aircraft.'}),
                    ('built', ('time', {}), {
                        'doc': 'The date the aircraft was constructed.'}),
                    ('make', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'The make of the aircraft.'}),
                    ('model', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'The model of the aircraft.'}),
                    ('serial', ('str', {'strip': True}), {
                        'doc': 'The serial number of the aircraft.'}),
                    ('operator', ('ps:contact', {}), {
                        'doc': 'Contact info representing the person or org that operates the aircraft.'}),
                )),
                ('transport:air:port', {}, (
                    ('name', ('str', {'lower': True, 'strip': True, 'onespace': True}), {
                        'doc': 'The name of the airport'}),
                    ('place', ('geo:place', {}), {
                        'doc': 'The place where the IATA airport code is assigned.'}),
                )),
                ('transport:air:tailnum', {}, (
                    ('loc', ('loc', {}), {
                        'doc': 'The geopolitical location that the tailnumber is allocated to.'}),
                    ('type', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'A type which may be specific to the country prefix.'}),
                )),
                ('transport:air:flightnum', {}, (
                    ('carrier', ('ou:org', {}), {
                        'doc': 'The org which operates the given flight number.'}),
                    ('to:port', ('transport:air:port', {}), {
                        'doc': 'The most recently registered destination for the flight number.'}),
                    ('from:port', ('transport:air:port', {}), {
                        'doc': 'The most recently registered origin for the flight number.'}),
                    ('stops', ('array', {'type': 'transport:air:port'}), {
                        'doc': 'An ordered list of aiport codes for the flight segments.'}),
                )),
                ('transport:air:flight', {}, (
                    ('num', ('transport:air:flightnum', {}), {
                        'doc': 'The flight number of this flight.'}),
                    ('scheduled:departure', ('time', {}), {
                        'doc': 'The time this flight was originally scheduled to depart'}),
                    ('scheduled:arrival', ('time', {}), {
                        'doc': 'The time this flight was originally scheduled to arrive'}),
                    ('departed', ('time', {}), {
                        'doc': 'The time this flight departed'}),
                    ('arrived', ('time', {}), {
                        'doc': 'The time this flight arrived'}),
                    ('carrier', ('ou:org', {}), {
                        'doc': 'The org which operates the given flight number.'}),
                    ('craft', ('transport:air:craft', {}), {
                        'doc': 'The aircraft that flew this flight.'}),
                    ('tailnum', ('transport:air:tailnum', {}), {
                        'doc': 'The tail/registration number at the time the aircraft flew this flight.'}),
                    ('to:port', ('transport:air:port', {}), {
                        'doc': 'The destination airport of this flight.'}),
                    ('from:port', ('transport:air:port', {}), {
                        'doc': 'The origin airport of this flight.'}),
                    ('stops', ('array', {'type': 'transport:air:port'}), {
                        'doc': 'An ordered list of aiport codes for stops which occured during this flight.'}),
                    ('cancelled', ('bool', {}), {
                        'doc': 'Set to true for cancelled flights.'}),
                )),
                ('transport:air:telem', {}, (
                    ('flight', ('transport:air:flight', {}), {
                        'doc': 'The flight being measured.'}),
                    ('latlong', ('geo:latlong', {}), {
                        'doc': 'The lat/lon of the aircraft at the time.'}),
                    ('loc', ('loc', {}), {
                        'doc': 'The location of the aircraft at the time.'}),
                    ('place', ('geo:place', {}), {
                        'doc': 'The place that the lat/lon geocodes to.'}),
                    ('accuracy', ('geo:dist', {}), {
                        'doc': 'The horizontal accuracy of the latlong sample.'}),
                    # TODO
                    # ('airspeed',
                    # ('groundspeed'
                    ('altitude', ('geo:altitude', {}), {
                        'doc': 'The altitude of the aircraft at the time.'}),
                    ('altitude:accuracy', ('geo:dist', {}), {
                        'doc': 'The vertical accuracy of the altitude measurement.'}),
                    ('time', ('time', {}), {
                        'doc': 'The time the telemetry sample was taken.'})
                )),
                ('transport:air:occupant', {}, (
                    ('type', ('str', {'lower': True}), {
                        'doc': 'The type of occupant such as pilot, crew or passenger.'}),
                    ('flight', ('transport:air:flight', {}), {
                        'doc': 'The flight that the occupant was aboard.'}),
                    ('seat', ('str', {'lower': True}), {
                        'doc': 'The seat assigned to the occupant'}),
                    ('contact', ('ps:contact', {}), {
                        'doc': 'The contact information of the occupant.'}),
                )),
                # TODO ais numbers
                ('transport:sea:vessel', {}, (
                    ('imo', ('transport:sea:imo', {}), {
                        'doc': 'The International Maritime Organization number for the vessel.'}),
                    ('name', ('str', {'lower': True, 'onespace': True, 'strip': True}), {
                        'doc': 'The name of the vessel'}),
                    ('length', ('geo:dist', {}), {
                        'doc': 'The official overall vessel length'}),
                    ('beam', ('geo:dist', {}), {
                        'doc': 'The official overall vessel beam'}),
                    ('flag', ('iso:3166:cc', {}), {
                        'doc': 'The country the vessel is flagged to.'}),
                    ('mmsi', ('transport:sea:mmsi', {}), {
                        'doc': 'The Maritime Mobile Service Identifier assigned to the vessel.'}),
                    ('built', ('time', {}), {
                        'doc': 'The year the vessel was constructed.'}),
                    ('operator', ('ps:contact', {}), {
                        'doc': 'The contact information of the operator.'}),
                    # TODO tonnage / gross tonnage?
                )),
                ('transport:sea:telem', {}, (
                    ('vessel', ('transport:sea:vessel', {}), {
                        'doc': 'The vessel being measured.'}),
                    ('time', ('time', {}), {
                        'doc': 'The time the telemetry was sampled.'}),
                    ('latlong', ('geo:latlong', {}), {
                        'doc': 'The lat/lon of the vessel at the time.'}),
                    ('loc', ('loc', {}), {
                        'doc': 'The location of the vessel at the time.'}),
                    ('place', ('geo:place', {}), {
                        'doc': 'The place that the lat/lon geocodes to.'}),
                    ('accuracy', ('geo:dist', {}), {
                        'doc': 'The horizontal accuracy of the latlong sample.'}),
                    ('draft', ('geo:dist', {}), {
                        'doc': 'The keel depth at the time.'}),
                    ('airdraft', ('geo:dist', {}), {
                        'doc': 'The maximum height of the ship from the waterline.'}),
                )),
            ),
        }
        return (('transport', modl), )



"""General XRPL Address Codec Exceptions."""
# It has been modified for vendored imports.
from synapse.vendor.xrpl.constants import XRPLException


class XRPLAddressCodecException(XRPLException):
    """General XRPL Address Codec Exception."""

    pass





# PE Resource Types
resource_types = (
    (1, 'RT_CURSOR'),
    (2, 'RT_BITMAP'),
    (3, 'RT_ICON'),
    (4, 'RT_MENU'),
    (5, 'RT_DIALOG'),
    (6, 'RT_STRING'),
    (7, 'RT_FONTDIR'),
    (8, 'RT_FONT'),
    (9, 'RT_ACCELERATOR'),
    (10, 'RT_RCDATA'),
    (11, 'RT_MESSAGETABLE'),
    (12, 'RT_GROUP_CURSOR'),
    (14, 'RT_GROUP_ICON'),
    (16, 'RT_VERSION'),
    (17, 'RT_DLGINCLUDE'),
    (19, 'RT_PLUGPLAY'),
    (20, 'RT_VXD'),
    (21, 'RT_ANICURSOR'),
    (22, 'RT_ANIICON'),
    (23, 'RT_HTML'),
    (24, 'RT_MANIFEST'),
)

def getRsrcTypes():
    return resource_types

# Language identifier constants and strings
# https://docs.microsoft.com/en-us/windows/desktop/intl/language-identifier-constants-and-strings#language-identifier-notes
langcodes = (
    (0x0C00, 'custom default'),  # Default custom locale language-Default custom sublanguage'
    (0x1400, 'ui_custom_default'),  # Default custom MUI locale language-Default custom MUI sublanguage
    (0x007f, 'invariant'),  # Invariant locale language-invariant sublanguage
    (0x0000, 'neutral'),  # Neutral locale language-Neutral sublanguage
    (0x0800, 'sys default'),  # System default locale language-System default sublanguage
    (0x1000, 'custom unspecified'),  # Unspecified custom locale language-Unspecified custom sublanguage
    (0x0400, 'default'),  # User default locale language-User default sublanguage
    (0x0436, 'af-ZA'),  # AFRIKAANS_SOUTH_AFRICA
    (0x041C, 'sq-AL'),  # ALBANIAN_ALBANIA
    (0x0484, 'gsw-FR'),  # ALSATIAN_FRANCE
    (0x045E, 'am-ET'),  # AMHARIC_ETHIOPIA
    (0x1401, 'ar-DZ'),  # ARABIC_ALGERIA
    (0x3C01, 'ar-BH'),  # ARABIC_BAHRAIN
    (0x0C01, 'ar-EG'),  # ARABIC_EGYPT
    (0x0801, 'ar-IQ'),  # ARABIC_IRAQ
    (0x2C01, 'ar-JO'),  # ARABIC_JORDAN
    (0x3401, 'ar-KW'),  # ARABIC_KUWAIT
    (0x3001, 'ar-LB'),  # ARABIC_LEBANON
    (0x1001, 'ar-LY'),  # ARABIC_LIBYA
    (0x1801, 'ar-MA'),  # ARABIC_MOROCCO
    (0x2001, 'ar-OM'),  # ARABIC_OMAN
    (0x4001, 'ar-QA'),  # ARABIC_QATAR
    (0x0401, 'ar-SA'),  # ARABIC_SAUDI_ARABIA
    (0x2801, 'ar-SY'),  # ARABIC_SYRIA
    (0x1C01, 'ar-TN'),  # ARABIC_TUNISIA
    (0x3801, 'ar-AE'),  # ARABIC_UAE
    (0x2401, 'ar-YE'),  # ARABIC_YEMEN
    (0x042B, 'hy-AM'),  # ARMENIAN_ARMENIA
    (0x044D, 'as-IN'),  # ASSAMESE_INDIA
    (0x082C, 'az-AZ-Cyrillic'),  # AZERI_CYRILLIC
    (0x042C, 'az-AZ-Latin'),  # AZERI_LATIN
    (0x0445, 'bn-Bangledesh'),  # BANGLA_BANGLADESH AND BANGLA_INDIA
    (0x0845, 'bn-IN'),  # BANGLA_INDIA
    (0x046D, 'ba-RU'),  # BASHKIR_RUSSIA
    (0x042D, 'Basque-Basque'),  # BASQUE_BASQUE
    (0x0423, 'be-BY'),  # BELARUSIAN_BELARUS
    (0x781A, 'bs-neutral'),  # BOSNIAN NEUTRAL
    (0x201A, 'bs-BA-Cyrillic'),  # BOSNIAN_BOSNIA_HERZEGOVINA_CYRILLIC
    (0x141A, 'bs-BA-Latin'),  # BOSNIAN_BOSNIA_HERZEGOVINA_LATIN
    (0x047E, 'br-FR'),  # BRETON_FRANCE
    (0x0402, 'bg-BG'),  # BULGARIAN_BULGARIA
    (0x0492, 'ku-IQ'),  # CENTRAL_KURDISH_IRAQ
    (0x045C, 'chr-Cher'),  # CHEROKEE_CHEROKEE
    (0x0403, 'ca-ES'),  # CATALAN_CATALAN
    (0x0C04, 'zh-HK'),  # CHINESE_HONGKONG
    (0x1404, 'zh-MO'),  # CHINESE_MACAU
    (0x1004, 'zh-SG'),  # CHINESE_SINGAPORE
    (0x0004, 'zh-Hans'),  # CHINESE_SIMPLIFIED
    (0x7C04, 'zh-Hant'),  # CHINESE_TRADITIONAL
    (0x0483, 'co-FR'),  # CORSICAN_FRANCE
    (0x001A, 'hr'),  # CROATIAN Neutral
    (0x101A, 'hr-BA'),  # CROATIAN_BOSNIA_HERZEGOVINA_LATIN
    (0x041A, 'hr-HR'),  # CROATIAN_CROATIA
    (0x0405, 'cs-CZ'),  # CZECH_CZECH_REPUBLIC
    (0x0406, 'da-DK'),  # DANISH_DENMARK
    (0x048C, 'prs-AF'),  # DARI_AFGHANISTAN
    (0x0465, 'dv-MV'),  # DIVEHI_MALDIVES
    (0x0813, 'nl-BE'),  # DUTCH_BELGIAN
    (0x0413, 'nl-NL'),  # DUTCH DUTCH
    (0x0C09, 'en-AU'),  # ENGLISH_AUS
    (0x2809, 'en-BZ'),  # ENGLISH_BELIZE
    (0x1009, 'en-CA'),  # ENGLISH_CAN
    (0x2409, 'en-029'),  # ENGLISH_CARIBBEAN
    (0x4009, 'en-IN'),  # ENGLISH_INDIA
    (0x1809, 'en-IE'),  # ENGLISH_IRELAND
    (0x2009, 'en-JM'),  # ENGLISH_JAMAICA
    (0x4409, 'en-MY'),  # ENGLISH_MALAYSIA
    (0x1409, 'en-NZ'),  # ENGLISH_NZ
    (0x3409, 'en-PH'),  # ENGLISH_PHILIPPINES
    (0x4809, 'en-SG'),  # ENGLISH_SINGAPORE
    (0x1c09, 'en-ZA'),  # ENGLISH_SOUTH_AFRICA
    (0x2C09, 'en-TT'),  # ENGLISH_TRINIDAD
    (0x0809, 'en-GB'),  # ENGLISH_UK
    (0x0409, 'en-US'),  # ENGLISH_US
    (0x3009, 'en-ZW'),  # ENGLISH_ZIMBABWE
    (0x0425, 'et-EE'),  # ESTONIAN_ESTONIA
    (0x0438, 'fo-FO'),  # FAEROESE_FAROE_ISLANDS
    (0x0464, 'fil-PH'),  # FILIPINO_PHILIPPINES
    (0x040B, 'fi-FI'),  # FINNISH_FINLAND
    (0x080c, 'fr-BE'),  # FRENCH_BELGIAN
    (0x0C0C, 'fr-CA'),  # FRENCH_CANADIAN
    (0x040c, 'fr-FR'),  # FRENCH_FRENCH
    (0x140C, 'fr-LU'),  # FRENCH_LUXEMBOURG
    (0x180C, 'fr-MC'),  # FRENCH_MONACO
    (0x100C, 'fr-CH'),  # FRENCH_SWISS
    (0x0462, 'fy-NL'),  # FRISIAN_NETHERLANDS
    (0x0456, 'gl-ES'),  # GALICIAN_GALICIAN
    (0x0437, 'ka-GE'),  # GEORGIAN_GEORGIA
    (0x0C07, 'de-AT'),  # GERMAN_AUSTRIAN
    (0x0407, 'de-DE'),  # GERMAN_GERMAN
    (0x1407, 'de-LI'),  # GERMAN_LIECHTENSTEIN
    (0x1007, 'de-LU'),  # GERMAN_LUXEMBOURG
    (0x0807, 'de-CH'),  # GERMAN_SWISS
    (0x0408, 'el-GR'),  # GREEK_GREECE
    (0x046F, 'kl-GL'),  # GREENLANDIC_GREENLAND
    (0x0447, 'gu-IN'),  # GUJARATI_INDIA
    (0x0468, 'ha-NG'),  # HAUSA_NIGERIA_LATIN
    (0x0475, 'haw-US'),  # HAWAIIAN_US
    (0x040D, 'he-IL'),  # HEBREW_ISRAEL
    (0x0439, 'hi-IN'),  # HINDI_INDIA
    (0x040E, 'hu-HU'),  # HUNGARIAN_HUNGARY
    (0x040F, 'is-IS'),  # ICELANDIC_ICELAND
    (0x0470, 'ig-NG'),  # IGBO_NIGERIA
    (0x0421, 'id-ID'),  # INDONESIAN_INDONESIA
    (0x085D, 'iu-CA-Latin'),  # INUKTITUT_CANADA_LATIN
    (0x045D, 'iu-CA'),  # INUKTITUT_CANADA
    (0x083C, 'ga-IE'),  # IRISH_IRELAND
    (0x0434, 'xh-ZA'),  # XHOSA_SOUTH_AFRICA
    (0x0435, 'zu-ZA'),  # ZULU_SOUTH_AFRICA
    (0x0410, 'it-IT'),  # ITALIAN_ITALIAN
    (0x0810, 'it-CH'),  # ITALIAN_SWISS
    (0x0411, 'ja-JP'),  # JAPANESE_JAPAN
    (0x044B, 'kn-IN'),  # KANNADA_INDIA
    (0x043F, 'kk-KZ'),  # KAZAK_KAZAKHSTAN
    (0x0453, 'kh-KH'),  # KHMER_CAMBODIA
    (0x0486, 'qut-GT'),  # KICHE_GUATEMALA
    (0x0487, 'rw-RW'),  # KINYARWANDA_RWANDA
    (0x0457, 'kok-IN'),  # KONKANI_INDIA
    (0x0412, 'ko-KR'),  # KOREA_KOREAN
    (0x0440, 'ky-KG'),  # KYRGYZ_KYRGYZSTAN
    (0x0454, 'lo-LA'),  # LAO_LAO
    (0x0426, 'lv-LV'),  # LATVIAN_LATVIA
    (0x0427, 'lt-LT'),  # LITHUANIAN_LITHUANIA
    (0x082E, 'dsb-DE'),  # LOWER_SORBIAN_GERMANY
    (0x046E, 'lb-LU'),  # LUXEMBOURGISH_LUXEMBOURG
    (0x042F, 'mk-MK'),  # MACEDONIAN_MACEDONIA
    (0x083E, 'ms-BN'),  # MALAY_BRUNEI_DARUSSALAM
    (0x043e, 'ms-MY'),  # MALAY_MALAYSIA
    (0x044C, 'ml-IN'),  # MALAYALAM_INDIA
    (0x043A, 'mt-MT'),  # MALTESE_MALTA
    (0x0481, 'mi-NZ'),  # MAORI_NEW_ZEALAND
    (0x047A, 'arn-CL'),  # MAPUDUNGUN_CHILE'
    (0x044E, 'mr-IN'),  # MARATHI_INDIA
    (0x047C, 'moh-CA'),  # MOHAWK_MOHAWK
    (0x0450, 'mn-MN-Cyrllic'),  # MONGOLIAN_CYRILLIC_MONGOLIA
    (0x0850, 'mn-MN-Prc'),  # MONGOLIAN_PRC
    (0x0461, 'ne-NP'),  # NEPALI_NEPAL
    (0x0414, 'nb-NO'),  # NORWEGIAN_BOKMAL
    (0x0814, 'no-NO'),  # NORWEGIAN_NYNORSK
    (0x0482, 'oc-FR'),  # OCCITAN_FRANCE
    (0x0448, 'or-IN'),  # ORIYA_INDIA
    (0x0463, 'ps-AF'),  # PASHTO_AFGHANISTAN
    (0x0429, 'fa-IR'),  # PERSIAN_IRAN
    (0x0415, 'pl-PL'),  # POLISH_POLAND
    (0x0416, 'pt-BR'),  # PORTUGUESE_BRAZILIAN
    (0x0816, 'pt-PT'),  # PORTUGUESE-PORTUGUESE
    (0x0867, 'ff-SN'),  # PULAR_SENEGAL
    (0x0446, 'pa-IN'),  # PUNJABI_INDIA
    (0x0846, 'pa-PK'),  # PUNJABI_PAKISTAN
    (0x046B, 'quz-BO'),  # QUECHUA_BOLIVIA
    (0x086B, 'quz-EC'),  # QUECHUA_ECUADOR
    (0x0C6B, 'quz-PE'),  # QUECHUA_PERU
    (0x0418, 'ro-RO'),  # ROMANIAN_ROMANIA
    (0x0417, 'rm-CH'),  # ROMANSH_SWITZERLAND
    (0x0419, 'ru-RU'),  # RUSSIAN_RUSSIA
    (0x0485, 'sah-RU'),  # SAKHA_RUSSIA
    (0x243B, 'smn-FIl'),  # SAMI_INARI_FINLAND
    (0x103B, 'smj-NO'),  # SAMI_LULE_NORWAY
    (0x143B, 'smj-SE'),  # SAMI_LULE_SWEDEN
    (0x0C3B, 'se-FI'),  # SAMI_NORTHERN_FINLAND
    (0x043B, 'se-NO'),  # SAMI_NORTHERN_NORWAY
    (0x083B, 'se-SE'),  # SAMI_NORTHERN_SWEDEN
    (0x203B, 'sms-FI'),  # SAMI_SKOLT_FINLAND
    (0x183B, 'sma-NO'),  # SAMI_SOUTHERN_NORWAY
    (0x1C3B, 'sma-SE'),  # SAMI_SOUTHERN_SWEDEN
    (0x044F, 'sa-IN'),  # SANSKRIT_INDIA
    (0x7C1A, 'sr-Neutral'),  # SERBIAN_NEUTRAL
    (0x1C1A, 'sr-BA'),  # SERBIAN_BOSNIA_HERZEGOVINA_CYRILLIC
    (0x181A, 'sr-code-Latin'),  # SERBIAN_BOSNIA_HERZEGOVINA_LATIN
    (0x0C1A, 'sr-CS-Cyrillic'),  # SERBIAN_CYRILLIC
    (0x081A, 'sr-CS-Latin'),  # SERBIAN_LATIN
    (0x046C, 'nso-ZA'),  # SOTHO_NORTHERN_SOUTH_AFRICA
    (0x0832, 'tn-BW'),  # TSWANA_BOTSWANA
    (0x0432, 'tn-ZA'),  # TSWANA_SOUTH_AFRICA
    (0x0859, 'sd-PK'),  # SINDHI_PAKISTAN
    (0x045B, 'si-LK'),  # SINHALESE_SRI_LANKA
    (0x041B, 'sk-SK'),  # SLOVAK_SLOVAKIA
    (0x0424, 'sl-SI'),  # SLOVENIAN_SLOVENIA
    (0x2C0A, 'es-AR'),  # SPANISH_ARGENTINA
    (0x400A, 'es-BO'),  # SPANISH_BOLIVIA
    (0x340A, 'es-CL'),  # SPANISH_CHILE
    (0x240A, 'es-CO'),  # SPANISH_COLOMBIA
    (0x140A, 'es-CR'),  # SPANISH_COSTA_RICA
    (0x1C0A, 'es-DO'),  # SPANISH_DOMINICAN_REPUBLIC
    (0x300A, 'es-EC'),  # SPANISH_ECUADOR
    (0x440A, 'es-SV'),  # SPANISH_EL_SALVADOR
    (0x100A, 'es-GT'),  # SPANISH_GUATEMALA
    (0x480A, 'es-HN'),  # SPANISH_HONDURAS
    (0x080A, 'es-MX'),  # SPANISH_MEXICAN
    (0x4C0A, 'es-NI'),  # SPANISH_NICARAGUA
    (0x180A, 'es-PA'),  # SPANISH_PANAMA
    (0x3C0A, 'es-PY'),  # SPANISH_PARAGUAY
    (0x280A, 'es-PE'),  # SPANISH_PERU
    (0x500A, 'es-PR'),  # SPANISH_PUERTO_RICO
    (0x0C0A, 'es-ES-modern'),  # SPANISH_MODERN
    (0x040A, 'es-ES-traditional'),  # SPANISH
    (0x540A, 'es-US'),  # SPANISH_US
    (0x380A, 'es-UY'),  # SPANISH_URUGUAY
    (0x200A, 'es-VE'),  # SPANISH_VENEZUELA
    (0x0441, 'sw-KE'),  # SWAHILI
    (0x081D, 'sv-FI'),  # SWEDISH_FINLAND
    (0x041D, 'sv-SE'),  # SWEDISH_SWEDEN
    (0x045A, 'syr-SY'),  # SYRIAC
    (0x0428, 'tg-TJ'),  # TAJIK_TAJIKISTAN
    (0x085F, 'tzm-DZ'),  # TAMAZIGHT_ALGERIA_LATIN
    (0x0449, 'ta-IN'),  # TAMIL_INDIA
    (0x0849, 'ta-LK'),  # TAMIL_SRI_LANKA
    (0x0444, 'tt-RU'),  # TATAR_RUSSIA
    (0x044A, 'te-IN'),  # TELUGU_INDIA
    (0x041E, 'th-TH'),  # THAI_THAILAND
    (0x0451, 'bo-CN'),  # TIBETAN_PRC
    (0x0873, 'ti-ER'),  # TIGRINYA_ERITREA
    (0x0473, 'ti-ET'),  # TIGRINYA_ETHIOPIA
    (0x041F, 'tr-TR'),  # TURKISH_TURKEY
    (0x0442, 'tk-TM'),  # TURKMEN_TURKMENISTAN
    (0x0422, 'uk-UA'),  # UKRAINIAN_UKRAINE
    (0x042E, 'hsb-DE'),  # UPPER_SORBIAN_GERMANY
    (0x0820, 'ur-IN'),  # URDU_INDIA
    (0x0420, 'ur-PK'),  # URDU_PAKISTAN
    (0x0480, 'ug-CN'),  # UIGHUR_PRC
    (0x0843, 'uz-UZ-Cyrillic'),  # UZBEK_CYRILLIC
    (0x0443, 'uz-UZ-Latin'),  # UZBEK_LATIN
    (0x0803, 'ca-ES-Valencia'),  # VALENCIAN_VALENCIA
    (0x042A, 'vi-VN'),  # VIETNAMESE_VIETNAM
    (0x0452, 'cy-GB'),  # WELSH_UNITED_KINGDOM
    (0x0488, 'wo-SN'),  # WOLOF_SENEGAL
    (0x0478, 'ii-CN'),  # YI_PRC
    (0x046A, 'yo-NG'),  # YORUBA_NIGERIA
)

def getLangCodes():
    return langcodes



import copy
import logging
import collections

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.chop as s_chop
import synapse.lib.time as s_time
import synapse.lib.layer as s_layer
import synapse.lib.stormtypes as s_stormtypes

logger = logging.getLogger(__name__)

class Node:
    '''
    A Cortex hypergraph node.

    NOTE: This object is for local Cortex use during a single Xact.
    '''
    def __init__(self, snap, sode, bylayer=None):
        self.snap = snap
        self.sode = sode

        self.buid = sode[0]

        # Tracks which property is retrieved from which layer
        self.bylayer = bylayer

        # if set, the node is complete.
        self.ndef = sode[1].get('ndef')
        self.form = snap.core.model.form(self.ndef[0])

        self.props = sode[1].get('props')
        if self.props is None:
            self.props = {}

        self.tags = sode[1].get('tags')
        if self.tags is None:
            self.tags = {}

        self.tagprops = sode[1].get('tagprops')
        if self.tagprops is None:
            self.tagprops = {}

        self.nodedata = sode[1].get('nodedata')
        if self.nodedata is None:
            self.nodedata = {}

    async def getStorNodes(self):
        '''
        Return a list of the raw storage nodes for each layer.
        '''
        return await self.snap.view.getStorNodes(self.buid)

    def getByLayer(self):
        '''
        Return a dictionary that translates the node's bylayer dict to a primitive.
        '''
        ndef = self.bylayer.get('ndef')
        tags = {t: l for (t, l) in self.bylayer.get('tags', {}).items()}
        props = {p: l for (p, l) in self.bylayer.get('props', {}).items()}
        tagprops = {p: l for (p, l) in self.bylayer.get('tagprops', {}).items()}
        return {'ndef': ndef, 'props': props, 'tags': tags, 'tagprops': tagprops}

    def __repr__(self):
        return f'Node{{{self.pack()}}}'

    async def addEdge(self, verb, n2iden):
        if self.form.isrunt:
            mesg = f'Edges cannot be used with runt nodes: {self.form.full}'
            raise s_exc.IsRuntForm(mesg=mesg, form=self.form.full)

        async with self.snap.getNodeEditor(self) as editor:
            return await editor.addEdge(verb, n2iden)

    async def delEdge(self, verb, n2iden):

        if not s_common.isbuidhex(n2iden):
            mesg = f'delEdge() got an invalid node iden: {n2iden}'
            raise s_exc.BadArg(mesg=mesg)

        nodeedits = (
            (self.buid, self.form.name, (
                (s_layer.EDIT_EDGE_DEL, (verb, n2iden), ()),
            )),
        )
        await self.snap.applyNodeEdits(nodeedits)

    async def iterEdgesN1(self, verb=None):
        async for edge in self.snap.iterNodeEdgesN1(self.buid, verb=verb):
            yield edge

    async def iterEdgesN2(self, verb=None):
        async for edge in self.snap.iterNodeEdgesN2(self.buid, verb=verb):
            yield edge

    async def storm(self, runt, text, opts=None, path=None):
        '''
        Args:
            path (Path):
                If set, then vars from path are copied into the new runtime, and vars are copied back out into path
                at the end

        Note:
            If opts is not None and opts['vars'] is set and path is not None, then values of path vars take precedent
        '''
        query = await self.snap.core.getStormQuery(text)

        if opts is None:
            opts = {}

        opts.setdefault('vars', {})
        if path is not None:
            opts['vars'].update(path.vars)

        async with runt.getSubRuntime(query, opts=opts) as subr:

            subr.addInput(self)

            async for subn, subp in subr.execute():
                yield subn, subp

            if path is not None:
                path.vars.update(subr.vars)

    async def filter(self, runt, text, opts=None, path=None):
        async for item in self.storm(runt, text, opts=opts, path=path):
            return False
        return True

    def iden(self):
        return s_common.ehex(self.buid)

    def pack(self, dorepr=False):
        '''
        Return the serializable/packed version of the node.

        Args:
            dorepr (bool): Include repr information for human readable versions of properties.

        Returns:
            (tuple): An (ndef, info) node tuple.
        '''

        node = (self.ndef, {
            'iden': self.iden(),
            'tags': self.tags,
            'props': self.props,
            'tagprops': self.tagprops,
            'nodedata': self.nodedata,
        })
        if dorepr:

            rval = self.repr()
            if rval is not None and rval != self.ndef[1]:
                node[1]['repr'] = self.repr()

            node[1]['reprs'] = self.reprs()
            node[1]['tagpropreprs'] = self.tagpropreprs()

        return node

    async def getEmbeds(self, embeds):
        '''
        Return a dictionary of property embeddings.
        '''
        retn = {}
        cache = {}
        async def walk(n, p):

            valu = n.props.get(p)
            if valu is None:
                return None

            prop = n.form.prop(p)
            if prop is None:
                return None

            if prop.modl.form(prop.type.name) is None:
                return None

            buid = s_common.buid((prop.type.name, valu))

            step = cache.get(buid, s_common.novalu)
            if step is s_common.novalu:
                step = cache[buid] = await node.snap.getNodeByBuid(buid)

            return step

        for nodepath, relprops in embeds.items():

            steps = nodepath.split('::')

            node = self
            for propname in steps:
                node = await walk(node, propname)
                if node is None:
                    break

            if node is None:
                continue

            embdnode = retn.get(nodepath)
            if embdnode is None:
                embdnode = retn[nodepath] = {}
                embdnode['*'] = s_common.ehex(node.buid)

            for relp in relprops:
                embdnode[relp] = node.props.get(relp)

        return retn

    async def seen(self, tick, source=None):
        '''
        Update the .seen interval and optionally a source specific seen node.
        '''
        await self.set('.seen', tick)

        if source is not None:
            seen = await self.snap.addNode('meta:seen', (source, self.ndef))
            await seen.set('.seen', tick)

    def getNodeRefs(self):
        '''
        Return a list of (prop, (form, valu)) refs out for the node.
        '''
        retn = []

        refs = self.form.getRefsOut()

        for name, dest in refs.get('prop', ()):
            valu = self.props.get(name)
            if valu is None:
                continue

            retn.append((name, (dest, valu)))

        for name in refs.get('ndef', ()):
            valu = self.props.get(name)
            if valu is None:
                continue
            retn.append((name, valu))

        for name, dest in refs.get('array', ()):

            valu = self.props.get(name)
            if valu is None:
                continue

            for item in valu:
                retn.append((name, (dest, item)))

        return retn

    async def set(self, name, valu, init=False):
        '''
        Set a property on the node.

        Args:
            name (str): The name of the property.
            valu (obj): The value of the property.
            init (bool): Set to True to disable read-only enforcement

        Returns:
            (bool): True if the property was changed.
        '''
        if self.snap.readonly:
            mesg = 'Cannot set property in read-only mode.'
            raise s_exc.IsReadOnly(mesg=mesg)

        prop = self.form.props.get(name)
        if prop is None:
            mesg = f'No property named {name} on form {self.form.name}.'
            await self.snap._raiseOnStrict(s_exc.NoSuchProp, mesg)
            return False

        if self.form.isrunt:
            if prop.info.get('ro'):
                mesg = 'Cannot set read-only props on runt nodes'
                raise s_exc.IsRuntForm(mesg=mesg, form=self.form.full, prop=name, valu=valu)

            await self.snap.core.runRuntPropSet(self, prop, valu)
            return True

        async with self.snap.getNodeEditor(self) as editor:
            return await editor.set(name, valu)

    def has(self, name):
        return name in self.props

    def get(self, name):
        '''
        Return a secondary property value from the Node.

        Args:
            name (str): The name of a secondary property.

        Returns:
            (obj): The secondary property value or None.
        '''
        if name.startswith('#'):
            return self.tags.get(name[1:])
        return self.props.get(name)

    async def _getPropDelEdits(self, name, init=False):

        prop = self.form.prop(name)
        if prop is None:
            if self.snap.strict:
                mesg = f'No property named {name}.'
                raise s_exc.NoSuchProp(mesg=mesg, name=name, form=self.form.name)
            await self.snap.warn(f'No Such Property: {name}')
            return ()

        if not init:

            if prop.info.get('ro'):
                if self.snap.strict:
                    raise s_exc.ReadOnlyProp(name=name)
                await self.snap.warn(f'Property is read-only: {name}')
                return ()

        curv = self.props.get(name, s_common.novalu)
        if curv is s_common.novalu:
            return ()

        edits = (
            (s_layer.EDIT_PROP_DEL, (prop.name, None, prop.type.stortype), ()),
        )
        return edits

    async def pop(self, name, init=False):
        '''
        Remove a property from a node and return the value
        '''
        if self.form.isrunt:
            prop = self.form.prop(name)
            if prop.info.get('ro'):
                raise s_exc.IsRuntForm(mesg='Cannot delete read-only props on runt nodes',
                                       form=self.form.full, prop=name)
            return await self.snap.core.runRuntPropDel(self, prop)

        edits = await self._getPropDelEdits(name, init=init)
        if not edits:
            return False

        await self.snap.applyNodeEdit((self.buid, self.form.name, edits))
        self.props.pop(name, None)
        return True

    def repr(self, name=None, defv=None):

        if name is None:
            return self.form.type.repr(self.ndef[1])

        prop = self.form.props.get(name)
        if prop is None:
            mesg = f'No property named {name}.'
            raise s_exc.NoSuchProp(mesg=mesg, form=self.form.name, prop=name)

        valu = self.props.get(name)
        if valu is None:
            return defv

        return prop.type.repr(valu)

    def reprs(self):
        '''
        Return a dictionary of repr values for props whose repr is different than
        the system mode value.
        '''
        reps = {}

        for name, valu in self.props.items():

            prop = self.form.prop(name)
            if prop is None:
                continue

            rval = prop.type.repr(valu)
            if rval is None or rval == valu:
                continue

            reps[name] = rval

        return reps

    def tagpropreprs(self):
        '''
        Return a dictionary of repr values for tagprops whose repr is different than
        the system mode value.
        '''
        reps = collections.defaultdict(dict)

        for tag, propdict in self.tagprops.items():
            for name, valu in propdict.items():

                prop = self.form.modl.tagprop(name)
                if prop is None:
                    continue

                rval = prop.type.repr(valu)
                if rval is None or rval == valu:
                    continue
                reps[tag][name] = rval

        return dict(reps)

    def hasTag(self, name):
        name = s_chop.tag(name)
        return name in self.tags

    def getTag(self, name, defval=None):
        name = s_chop.tag(name)
        return self.tags.get(name, defval)

    def getTags(self, leaf=False):

        if not leaf:
            return list(self.tags.items())

        # longest first
        retn = []

        # brute force rather than build a tree.  faster in small sets.
        for _, tag, valu in sorted([(len(t), t, v) for (t, v) in self.tags.items()], reverse=True):

            look = tag + '.'
            if any([r.startswith(look) for (r, rv) in retn]):
                continue

            retn.append((tag, valu))

        return retn

    async def addTag(self, tag, valu=(None, None)):
        '''
        Add a tag to a node.

        Args:
            tag (str): The tag to add to the node.
            valu: The optional tag value.  If specified, this must be a value that
                  norms as a valid time interval as an ival.

        Returns:
            None: This returns None.
        '''
        if self.form.isrunt:
            raise s_exc.IsRuntForm(mesg='Cannot add tags to runt nodes.',
                                   form=self.form.full, tag=tag)

        async with self.snap.getNodeEditor(self) as protonode:
            await protonode.addTag(tag, valu=valu)

    def _getTagTree(self):

        root = (None, {})
        for tag in self.tags.keys():

            node = root

            for part in tag.split('.'):

                kidn = node[1].get(part)

                if kidn is None:

                    full = part
                    if node[0] is not None:
                        full = f'{node[0]}.{full}'

                    kidn = node[1][part] = (full, {})

                node = kidn

        return root

    async def _getTagDelEdits(self, tag, init=False):

        path = s_chop.tagpath(tag)

        name = '.'.join(path)

        if self.form.isrunt:
            raise s_exc.IsRuntForm(mesg='Cannot delete tags from runt nodes.',
                                   form=self.form.full, tag=tag)

        curv = self.tags.get(name, s_common.novalu)
        if curv is s_common.novalu:
            return ()

        pref = name + '.'

        todel = [(len(t), t) for t in self.tags.keys() if t.startswith(pref)]

        if len(path) > 1:

            parent = '.'.join(path[:-1])

            # retrieve a list of prunable tags
            prune = await self.snap.core.getTagPrune(parent)
            if prune:

                tree = self._getTagTree()

                for prunetag in reversed(prune):

                    node = tree
                    for step in prunetag.split('.'):

                        node = node[1].get(step)
                        if node is None:
                            break

                    if node is not None and len(node[1]) == 1:
                        todel.append((len(node[0]), node[0]))
                        continue

                    break

        todel.sort(reverse=True)

        # order matters...
        edits = []

        for _, subtag in todel:

            edits.extend(self._getTagPropDel(subtag))
            edits.append((s_layer.EDIT_TAG_DEL, (subtag, None), ()))

        edits.extend(self._getTagPropDel(name))
        edits.append((s_layer.EDIT_TAG_DEL, (name, None), ()))

        return edits

    async def delTag(self, tag, init=False):
        '''
        Delete a tag from the node.
        '''
        edits = await self._getTagDelEdits(tag, init=init)
        if edits:
            nodeedit = (self.buid, self.form.name, edits)
            await self.snap.applyNodeEdit(nodeedit)

    def _getTagPropDel(self, tag):

        edits = []
        for tagprop in self.getTagProps(tag):

            prop = self.snap.core.model.getTagProp(tagprop)

            if prop is None:  # pragma: no cover
                logger.warn(f'Cant delete tag prop ({tagprop}) without model prop!')
                continue
            edits.append((s_layer.EDIT_TAGPROP_DEL, (tag, tagprop, None, prop.type.stortype), ()))

        return edits

    def getTagProps(self, tag):
        propdict = self.tagprops.get(tag)
        if not propdict:
            return []
        return list(propdict.keys())

    def hasTagProp(self, tag, prop):
        '''
        Check if a #foo.bar:baz tag property exists on the node.
        '''
        return tag in self.tagprops and prop in self.tagprops[tag]

    def getTagProp(self, tag, prop, defval=None):
        '''
        Return the value (or defval) of the given tag property.
        '''
        propdict = self.tagprops.get(tag)
        if propdict:
            return propdict.get(prop, defval)
        return defval

    async def setTagProp(self, tag, name, valu):
        '''
        Set the value of the given tag property.
        '''
        async with self.snap.getNodeEditor(self) as editor:
            await editor.setTagProp(tag, name, valu)

    async def delTagProp(self, tag, name):
        prop = self.snap.core.model.getTagProp(name)
        if prop is None:
            raise s_exc.NoSuchTagProp(name=name)

        propdict = self.tagprops.get(tag)
        if not propdict:
            return False

        curv = propdict.get(name, s_common.novalu)
        if curv is s_common.novalu:
            return False

        edits = (
            (s_layer.EDIT_TAGPROP_DEL, (tag, name, None, prop.type.stortype), ()),
        )

        await self.snap.applyNodeEdit((self.buid, self.form.name, edits))

    async def delete(self, force=False):
        '''
        Delete a node from the cortex.

        The following tear-down operations occur in order:

            * validate that you have permissions to delete the node
            * validate that you have permissions to delete all tags
            * validate that there are no remaining references to the node.

            * delete all the tags (bottom up)
                * fire onDelTag() handlers
                * delete tag properties from storage
                * log tag:del splices

            * delete all secondary properties
                * fire onDelProp handler
                * delete secondary property from storage
                * log prop:del splices

            * delete the primary property
                * fire onDel handlers for the node
                * delete primary property from storage
                * log node:del splices
        '''

        formname, formvalu = self.ndef

        if self.form.isrunt:
            raise s_exc.IsRuntForm(mesg='Cannot delete runt nodes',
                                   form=formname, valu=formvalu)

        # top level tags will cause delete cascades
        tags = [t for t in self.tags.keys() if len(t.split('.')) == 1]

        # check for any nodes which reference us...
        if not force:

            # refuse to delete tag nodes with existing tags
            if self.form.name == 'syn:tag':

                async for _ in self.snap.nodesByTag(self.ndef[1]):  # NOQA
                    mesg = 'Nodes still have this tag.'
                    return await self.snap._raiseOnStrict(s_exc.CantDelNode, mesg, form=formname,
                                                          iden=self.iden())

            async for refr in self.snap.nodesByPropTypeValu(formname, formvalu):

                if refr.buid == self.buid:
                    continue

                mesg = 'Other nodes still refer to this node.'
                return await self.snap._raiseOnStrict(s_exc.CantDelNode, mesg, form=formname,
                                                      iden=self.iden())

        edits = []
        for tag in tags:
            edits.extend(await self._getTagDelEdits(tag, init=True))

        for name in self.props.keys():
            edits.extend(await self._getPropDelEdits(name, init=True))

        edits.append(
            (s_layer.EDIT_NODE_DEL, (formvalu, self.form.type.stortype), ()),
        )

        await self.snap.applyNodeEdit((self.buid, formname, edits))
        self.snap.livenodes.pop(self.buid, None)

    async def hasData(self, name):
        if name in self.nodedata:
            return True
        return await self.snap.hasNodeData(self.buid, name)

    async def getData(self, name, defv=None):
        valu = self.nodedata.get(name, s_common.novalu)
        if valu is not s_common.novalu:
            return valu
        return await self.snap.getNodeData(self.buid, name, defv=defv)

    async def setData(self, name, valu):
        async with self.snap.getNodeEditor(self) as protonode:
            await protonode.setData(name, valu)

    async def popData(self, name):
        retn = await self.snap.getNodeData(self.buid, name)

        edits = (
            (s_layer.EDIT_NODEDATA_DEL, (name, None), ()),
        )
        await self.snap.applyNodeEdits(((self.buid, self.form.name, edits),))

        return retn

    async def iterData(self):
        async for item in self.snap.iterNodeData(self.buid):
            yield item

class Path:
    '''
    A path context tracked through the storm runtime.
    '''
    def __init__(self, vars, nodes):

        self.node = None
        self.nodes = nodes

        if len(nodes):
            self.node = nodes[-1]

        self.vars = vars
        self.frames = []
        self.ctors = {}

        # "builtins" which are *not* vars
        # ( this allows copying variable context )
        self.builtins = {
            'path': self,
            'node': self.node,
        }

        self.metadata = {}

    def getVar(self, name, defv=s_common.novalu):

        # check if the name is in our variables
        valu = self.vars.get(name, s_common.novalu)
        if valu is not s_common.novalu:
            return valu

        # check if it's in builtins
        valu = self.builtins.get(name, s_common.novalu)
        if valu is not s_common.novalu:
            return valu

        ctor = self.ctors.get(name)
        if ctor is not None:
            valu = ctor(self)
            self.vars[name] = valu
            return valu

        return s_common.novalu

    async def setVar(self, name, valu):
        self.vars[name] = valu

    async def popVar(self, name):
        return self.vars.pop(name, s_common.novalu)

    def meta(self, name, valu):
        '''
        Add node specific metadata to be returned with the node.
        '''
        self.metadata[name] = valu

    async def pack(self, path=False):
        ret = dict(self.metadata)
        if ret:
            ret = await s_stormtypes.toprim(ret)
        if path:
            ret['nodes'] = [node.iden() for node in self.nodes]
        return ret

    def fork(self, node):

        nodes = list(self.nodes)
        nodes.append(node)

        path = Path(self.vars.copy(), nodes)

        return path

    def clone(self):
        path = Path(copy.copy(self.vars), copy.copy(self.nodes))
        path.frames = [v.copy() for v in self.frames]
        return path

    def initframe(self, initvars=None):

        framevars = {}
        if initvars is not None:
            framevars.update(initvars)

        self.frames.append(self.vars)

        self.vars = framevars

    def finiframe(self):
        '''
        Pop a scope frame from the path, restoring runt if at the top
        Args:
            runt (Runtime): A storm runtime to restore if we're at the top
            merge (bool): Set to true to merge vars back up into the next frame
        '''
        if not self.frames:
            self.vars.clear()
            return

        self.vars = self.frames.pop()

def props(pode):
    '''
    Get the props from the node.

    Args:
        pode (tuple): A packed node.

    Notes:
        This will include any universal props present on the node.

    Returns:
        dict: A dictionary of properties.
    '''
    return pode[1]['props'].copy()

def prop(pode, prop):
    '''
    Return the valu of a given property on the node.

    Args:
        pode (tuple): A packed node.
        prop (str): Property to retrieve.

    Notes:
        The prop argument may be the full property name (foo:bar:baz), relative property name (:baz) , or the unadorned
        property name (baz).

    Returns:

    '''
    form = pode[0][0]
    if prop.startswith(form):
        prop = prop[len(form):]
    if prop[0] == ':':
        prop = prop[1:]
    return pode[1]['props'].get(prop)

def tags(pode, leaf=False):
    '''
    Get all the tags for a given node.

    Args:
        pode (tuple): A packed node.
        leaf (bool): If True, only return leaf tags

    Returns:
        list: A list of tag strings.
    '''
    if not leaf:
        return list(pode[1]['tags'].keys())
    return _tagscommon(pode, True)

def tagsnice(pode):
    '''
    Get all the leaf tags and the tags that have values or tagprops.

    Args:
        pode (tuple): A packed node.

    Returns:
        list: A list of tag strings.
    '''
    ret = _tagscommon(pode, False)
    for tag in pode[1].get('tagprops', {}):
        if tag not in ret:
            ret.append(tag)
    return ret

def _tagscommon(pode, leafonly):
    '''
    Return either all the leaf tags or all the leaf tags and all the internal tags with values
    '''
    retn = []

    # brute force rather than build a tree.  faster in small sets.
    for tag, val in sorted((t for t in pode[1]['tags'].items()), reverse=True, key=lambda x: len(x[0])):
        look = tag + '.'
        if (leafonly or val == (None, None)) and any([r.startswith(look) for r in retn]):
            continue
        retn.append(tag)
    return retn

def tagged(pode, tag):
    '''
    Check if a packed node has a given tag.

    Args:
        pode (tuple): A packed node.
        tag (str): The tag to check.

    Examples:
        Check if a node is tagged with "woot" and dostuff if it is.

            if s_node.tagged(node,'woot'):
                dostuff()

    Notes:
        If the tag starts with `#`, this is removed prior to checking.

    Returns:
        bool: True if the tag is present. False otherwise.
    '''
    if tag.startswith('#'):
        tag = tag[1:]
    return pode[1]['tags'].get(tag) is not None

def ndef(pode):
    '''
    Return a node definition (<form>,<valu>) tuple from the node.

    Args:
        pode (tuple): A packed node.

    Returns:
        ((str,obj)):    The (<form>,<valu>) tuple for the node
    '''
    return pode[0]

def iden(pode):
    '''
    Return the iden (buid) of the packed node.

    Args:
        pode (tuple): A packed node.

    Returns:
        str: The node iden.
    '''
    return pode[1].get('iden')

def reprNdef(pode):
    '''
    Get the ndef of the pode with a human readable value.

    Args:
        pode (tuple): A packed node.

    Notes:
        The human readable value is only available if the node came from a
        storm query execution where the ``repr`` key was passed into the
        ``opts`` argument with a True value.

    Returns:
        (str, str): A tuple of form and the human readable value.

    '''
    ((form, valu), info) = pode
    formvalu = info.get('repr')
    if formvalu is None:
        formvalu = str(valu)
    return form, formvalu

def reprProp(pode, prop):
    '''
    Get the human readable value for a secondary property from the pode.

    Args:
        pode (tuple): A packed node.
        prop:

    Notes:
        The human readable value is only available if the node came from a
        storm query execution where the ``repr`` key was passed into the
        ``opts`` argument with a True value.

        The prop argument may be the full property name (foo:bar:baz), relative
        property name (:baz) , or the unadorned property name (baz).

    Returns:
        str: The human readable property value.  If the property is not present, returns None.
    '''
    form = pode[0][0]
    if prop.startswith(form):
        prop = prop[len(form):]
    if prop[0] == ':':
        prop = prop[1:]
    opropvalu = pode[1].get('props').get(prop)
    if opropvalu is None:
        return None
    propvalu = pode[1].get('reprs', {}).get(prop)
    if propvalu is None:
        return str(opropvalu)
    return propvalu

def reprTag(pode, tag):
    '''
    Get the human readable value for the tag timestamp from the pode.

    Args:
        pode (tuple): A packed node.
        tag (str): The tag to get the value for.

    Notes:
        The human readable value is only available if the node came from a
        storm query execution where the ``repr`` key was passed into the
        ``opts`` argument with a True value.

        If the tag does not have a timestamp, this returns a empty string.

    Returns:
        str: The human readable value for the tag. If the tag is not present, returns None.
    '''
    tag = tag.lstrip('#')
    valu = pode[1]['tags'].get(tag)
    if valu is None:
        return None
    if valu == (None, None):
        return ''
    mint = s_time.repr(valu[0])
    maxt = s_time.repr(valu[1])
    valu = f'({mint}, {maxt})'
    return valu

def reprTagProps(pode, tag):
    '''
    Get the human readable values for any tagprops on a tag for a given node.

    Args:
        pode (tuple): A packed node.
        tag (str): The tag to get the tagprops reprs for.

    Notes:
        The human readable value is only available if the node came from a
        storm query execution where the ``repr`` key was passed into the
        ``opts`` argument with a True value.

        If the tag does not have any tagprops associated with it, this returns an empty list.

    Returns:
        list: A list of tuples, containing the name of the tagprop and the repr value.
    '''
    ret = []
    exists = pode[1]['tags'].get(tag)
    if exists is None:
        return ret
    tagprops = pode[1].get('tagprops', {}).get(tag)
    if tagprops is None:
        return ret
    for prop, valu in tagprops.items():
        rval = pode[1].get('tagpropreprs', {}).get(tag, {}).get(prop)
        if rval is not None:
            ret.append((prop, rval))
        else:
            ret.append((prop, str(valu)))
    return sorted(ret, key=lambda x: x[0])



import synapse.lib.module as s_module

class GovIntlModule(s_module.CoreModule):

    def getModelDefs(self):
        modl = {
            'types': (
                ('iso:oid', ('str', {'regex': '^([0-2])((\\.0)|(\\.[1-9][0-9]*))*$'}), {
                    'doc': 'An ISO Object Identifier string.'}),

                ('iso:3166:cc', ('str', {'lower': True, 'regex': '^[a-z]{2}$'}), {
                    'doc': 'An ISO 3166 2 digit country code.'}),

                ('gov:intl:un:m49', ('int', {'min': 1, 'max': 999}), {
                    'doc': 'UN M49 Numeric Country Code.'}),
            ),

            'forms': (
                ('iso:oid', {}, (
                    ('descr', ('str', {}), {
                        'doc': 'A description of the value or meaning of the OID.'}),
                    ('identifier', ('str', {}), {
                        'doc': 'The string identifier for the deepest tree element.'}),
                )),
            ),
        }
        name = 'gov:intl'
        return ((name, modl), )



import asyncio
import logging

import synapse.exc as s_exc
import synapse.lib.stormtypes as s_stormtypes

logger = logging.getLogger(__name__)


storm_missing_autoadds = '''
$absoluteOrder = $lib.view.list(deporder=$lib.true)

if $lib.debug {
    $lib.print('The following Views will be fixed in order:')
    for $view in $absoluteOrder {
        $lib.print($view.iden)
    }
}

$queries = ( ${ inet:dns:request:query:name:fqdn $valu=:query:name:fqdn [inet:fqdn=$valu] },
${ inet:dns:request:query:name:ipv4 $valu=:query:name:ipv4 [inet:ipv4=$valu] },
${ inet:dns:request:query:name:ipv6 $valu=:query:name:ipv6 [inet:ipv6=$valu] },
${ inet:dns:query:name:fqdn $valu=:name:fqdn [inet:fqdn=$valu] },
${ inet:dns:query:name:ipv4 $valu=:name:ipv4 [inet:ipv4=$valu] },
${ inet:dns:query:name:ipv6 $valu=:name:ipv6 [inet:ipv6=$valu] },
${ inet:asnet4:net4:min $valu=:net4:min [inet:ipv4=$valu] },
${ inet:asnet4:net4:max $valu=:net4:max [inet:ipv4=$valu] },
${ inet:asnet6:net6:min $valu=:net6:min [inet:ipv6=$valu] },
${ inet:asnet6:net6:max $valu=:net6:max [inet:ipv6=$valu] },
${ inet:whois:iprec:net4:min $valu=:net4:min [inet:ipv4=$valu] },
${ inet:whois:iprec:net4:max $valu=:net4:max [inet:ipv4=$valu] },
${ inet:whois:iprec:net6:min $valu=:net6:min [inet:ipv6=$valu] },
${ inet:whois:iprec:net6:max $valu=:net6:max [inet:ipv6=$valu] },
${ it:app:snort:hit:src:ipv4 $valu=:src:ipv4 [inet:ipv4=$valu] },
${ it:app:snort:hit:src:ipv6 $valu=:src:ipv6 [inet:ipv6=$valu] },
${ it:app:snort:hit:dst:ipv4 $valu=:dst:ipv4 [inet:ipv4=$valu] },
${ it:app:snort:hit:dst:ipv6 $valu=:dst:ipv6 [inet:ipv6=$valu] },)

for $view in $absoluteOrder {
    if $lib.debug { $lib.print('Fixing autoadd data in view {v}', v=$view) }
    for $query in $queries {
        if $lib.debug { $lib.print('Executing { {q} }', q=$query) }
        view.exec $view.iden $query
    }
}
'''

storm_missing_coins = '''
    for $view in $lib.view.list(deporder=$lib.true) {
        view.exec $view.iden {
            $coins = $lib.set()
            crypto:currency:address
            $coins.add(:coin) | spin |
            for $coin in $coins {[ crypto:currency:coin=$coin ]}
        }
    }
'''

storm_missing_cpe22 = '''
$views = $lib.view.list(deporder=$lib.true)
for $view in $views {
    view.exec $view.iden { it:sec:cpe -:v2_2 [ :v2_2=$node.value() ] }
}
'''


hotfixes = (
    ((1, 0, 0), {
        'desc': 'Create nodes for known missing autoadds.',
        'query': storm_missing_autoadds,
    }),
    ((2, 0, 0), {
        'desc': 'Populate crypto:currency:coin nodes from existing addresses.',
        'query': storm_missing_coins,
    }),
    ((3, 0, 0), {
        'desc': 'Populate it:sec:cpe:v2_2 properties from existing CPE where the property is not set.',
        'query': storm_missing_cpe22,
    }),
)
runtime_fixes_key = 'cortex:runtime:stormfixes'

def getMaxHotFixes():
    return max([vers for vers, info in hotfixes])

@s_stormtypes.registry.registerLib
class CellLib(s_stormtypes.Lib):
    '''
    A Storm Library for interacting with the Cortex.
    '''
    _storm_locals = (
        {'name': 'getCellInfo', 'desc': 'Return metadata specific for the Cortex.',
         'type': {'type': 'function', '_funcname': '_getCellInfo', 'args': (),
                  'returns': {'type': 'dict', 'desc': 'A dictionary containing metadata.', }}},
        {'name': 'getBackupInfo', 'desc': 'Get information about recent backup activity.',
         'type': {'type': 'function', '_funcname': '_getBackupInfo', 'args': (),
                  'returns': {'type': 'dict', 'desc': 'A dictionary containing backup information.', }}},
        {'name': 'getSystemInfo', 'desc': 'Get info about the system in which the Cortex is running.',
         'type': {'type': 'function', '_funcname': '_getSystemInfo', 'args': (),
                  'returns': {'type': 'dict', 'desc': 'A dictionary containing system information.', }}},
        {'name': 'getHealthCheck', 'desc': 'Get healthcheck information about the Cortex.',
         'type': {'type': 'function', '_funcname': '_getHealthCheck', 'args': (),
                  'returns': {'type': 'dict', 'desc': 'A dictionary containing healthcheck information.', }}},
        {'name': 'hotFixesApply', 'desc': 'Apply known data migrations and fixes via storm.',
         'type': {'type': 'function', '_funcname': '_hotFixesApply', 'args': (),
                  'returns': {'type': 'list',
                              'desc': 'Tuple containing the current version after applying the fixes.'}}},
        {'name': 'hotFixesCheck', 'desc': 'Check to see if there are known hot fixes to apply.',
         'type': {'type': 'function', '_funcname': '_hotFixesCheck', 'args': (),
                  'returns': {'type': 'boolean', 'desc': 'Bool indicating if there are hot fixes to apply or not.'}}},
        {'name': 'trimNexsLog', 'desc': '''
            Rotate and cull the Nexus log (and any consumers) at the current offset.

            If the consumers argument is provided they will first be checked
            if online before rotating and raise otherwise.
            After rotation, all consumers provided must catch-up to the offset to cull at
            within the specified timeout before executing the cull, and will raise otherwise.
         ''',
         'type': {'type': 'function', '_funcname': '_trimNexsLog',
                  'args': (
                      {'name': 'consumers', 'type': 'array', 'default': None,
                       'desc': 'List of Telepath URLs for consumers of the Nexus log.'},
                      {'name': 'timeout', 'type': 'int', 'default': 30,
                       'desc': 'Time (in seconds) to wait for consumers to catch-up before culling.'}
                  ),
                  'returns': {'type': 'int', 'desc': 'The offset that was culled (up to and including).'}}},
    )
    _storm_lib_path = ('cell',)

    def getObjLocals(self):
        return {
            'getCellInfo': self._getCellInfo,
            'getBackupInfo': self._getBackupInfo,
            'getSystemInfo': self._getSystemInfo,
            'getHealthCheck': self._getHealthCheck,
            'hotFixesApply': self._hotFixesApply,
            'hotFixesCheck': self._hotFixesCheck,
            'trimNexsLog': self._trimNexsLog,
        }

    async def _hotFixesApply(self):
        if not self.runt.isAdmin():
            mesg = '$lib.cell.stormFixesApply() requires admin privs.'
            raise s_exc.AuthDeny(mesg=mesg)

        curv = await self.runt.snap.core.getStormVar(runtime_fixes_key, default=(0, 0, 0))
        for vers, info in hotfixes:
            if vers <= curv:
                continue

            desc = info.get('desc')
            text = info.get('query')
            vars = info.get('vars', {})
            assert text is not None
            assert desc is not None
            assert vars is not None

            await self.runt.printf(f'Applying hotfix {vers} for [{desc}]')
            try:
                query = await self.runt.getStormQuery(text)
                async with self.runt.getSubRuntime(query, opts={'vars': vars}) as runt:
                    async for item in runt.execute():
                        pass
            except asyncio.CancelledError: # pragma: no cover
                raise
            except Exception as e: # pragma: no cover
                logger.exception(f'Error applying storm hotfix {vers}')
                raise
            else:
                await self.runt.snap.core.setStormVar(runtime_fixes_key, vers)
                await self.runt.printf(f'Applied hotfix {vers}')
            curv = vers

        return curv

    async def _hotFixesCheck(self):
        if not self.runt.isAdmin():
            mesg = '$lib.cell.stormFixesCheck() requires admin privs.'
            raise s_exc.AuthDeny(mesg=mesg)

        curv = await self.runt.snap.core.getStormVar(runtime_fixes_key, default=(0, 0, 0))

        dowork = False
        for vers, info in hotfixes:
            if vers <= curv:
                continue

            dowork = True
            desc = info.get('desc')
            await self.runt.printf(f'Would apply fix {vers} for [{desc}]')

        return dowork

    async def _getCellInfo(self):
        if not self.runt.isAdmin():
            mesg = '$lib.cell.getCellInfo() requires admin privs.'
            raise s_exc.AuthDeny(mesg=mesg)
        return await self.runt.snap.core.getCellInfo()

    async def _getSystemInfo(self):
        if not self.runt.isAdmin():
            mesg = '$lib.cell.getSystemInfo() requires admin privs.'
            raise s_exc.AuthDeny(mesg=mesg)
        return await self.runt.snap.core.getSystemInfo()

    async def _getBackupInfo(self):
        if not self.runt.isAdmin():
            mesg = '$lib.cell.getBackupInfo() requires admin privs.'
            raise s_exc.AuthDeny(mesg=mesg)
        return await self.runt.snap.core.getBackupInfo()

    async def _getHealthCheck(self):
        if not self.runt.isAdmin():
            mesg = '$lib.cell.getHealthCheck() requires admin privs.'
            raise s_exc.AuthDeny(mesg=mesg)
        return await self.runt.snap.core.getHealthCheck()

    async def _trimNexsLog(self, consumers=None, timeout=30):
        if not self.runt.isAdmin():
            mesg = '$lib.cell.trimNexsLog() requires admin privs.'
            raise s_exc.AuthDeny(mesg=mesg)

        timeout = await s_stormtypes.toint(timeout, noneok=True)

        if consumers is not None:
            consumers = [await s_stormtypes.tostr(turl) async for turl in s_stormtypes.toiter(consumers)]

        return await self.runt.snap.core.trimNexsLog(consumers=consumers, timeout=timeout)



import os
import asyncio
import logging

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.cell as s_cell
import synapse.lib.coro as s_coro
import synapse.lib.nexus as s_nexus
import synapse.lib.jsonstor as s_jsonstor
import synapse.lib.lmdbslab as s_lmdbslab

logger = logging.getLogger(__name__)

class AhaApi(s_cell.CellApi):

    async def getAhaUrls(self):
        return self.cell.conf.get('aha:urls', ())

    async def getAhaSvc(self, name):
        '''
        Return an AHA service description dictionary for a fully qualified service name.
        '''
        svcinfo = await self.cell.getAhaSvc(name)
        if svcinfo is None:
            return None

        svcnetw = svcinfo.get('svcnetw')
        await self._reqUserAllowed(('aha', 'service', 'get', svcnetw))
        return svcinfo

    async def getAhaSvcs(self, network=None):
        '''
        Yield AHA svcinfo dictionaries.

        Args:
            network (str): Optionally specify a network to filter on.
        '''
        if network is None:
            await self._reqUserAllowed(('aha', 'service', 'get'))
        else:
            await self._reqUserAllowed(('aha', 'service', 'get', network))

        async for info in self.cell.getAhaSvcs(network=network):
            yield info

    async def addAhaSvc(self, name, info, network=None):
        '''
        Register a service with the AHA discovery server.

        NOTE: In order for the service to remain marked "up" a caller
              must maintain the telepath link.
        '''
        svcname, svcnetw, svcfull = self.cell._nameAndNetwork(name, network)

        await self._reqUserAllowed(('aha', 'service', 'add', svcnetw, svcname))

        # dont disclose the real session...
        sess = s_common.guid(self.sess.iden)
        info['online'] = sess

        if self.link.sock is not None:
            host, port = self.link.sock.getpeername()
            urlinfo = info.get('urlinfo', {})
            urlinfo.setdefault('host', host)

        async def fini():
            if self.cell.isfini:  # pragma: no cover
                mesg = f'{self.cell.__class__.__name__} is fini. Unable to set {name}@{network} as down.'
                logger.warning(mesg)
                return

            logger.debug(f'AhaCellApi fini, tearing down [{name}]')
            coro = self.cell.setAhaSvcDown(name, sess, network=network)
            self.cell.schedCoro(coro)  # this will eventually execute or get cancelled.

        self.onfini(fini)

        return await self.cell.addAhaSvc(name, info, network=network)

    async def delAhaSvc(self, name, network=None):
        '''
        Remove an AHA service entry.
        '''
        svcname, svcnetw, svcfull = self.cell._nameAndNetwork(name, network)
        await self._reqUserAllowed(('aha', 'service', 'del', svcnetw, svcname))
        return await self.cell.delAhaSvc(name, network=network)

    async def getCaCert(self, network):

        await self._reqUserAllowed(('aha', 'ca', 'get'))
        return await self.cell.getCaCert(network)

    async def genCaCert(self, network):

        await self._reqUserAllowed(('aha', 'ca', 'gen'))
        return await self.cell.genCaCert(network)

    async def signHostCsr(self, csrtext, signas=None, sans=None):

        await self._reqUserAllowed(('aha', 'csr', 'host'))
        return await self.cell.signHostCsr(csrtext, signas=signas, sans=sans)

    async def signUserCsr(self, csrtext, signas=None):

        await self._reqUserAllowed(('aha', 'csr', 'user'))
        return await self.cell.signUserCsr(csrtext, signas=signas)

class AhaCell(s_cell.Cell):

    cellapi = AhaApi
    confdefs = {
        'aha:urls': {
            'description': 'A list of all available AHA server URLs.',
            'type': ['string', 'array'],
            'items': {'type': 'string'},
        },
    }

    async def initServiceStorage(self):

        # TODO plumb using a remote jsonstor?
        dirn = s_common.gendir(self.dirn, 'slabs', 'jsonstor')

        slab = await s_lmdbslab.Slab.anit(dirn)
        self.jsonstor = await s_jsonstor.JsonStor.anit(slab, 'aha')  # type: s_jsonstor.JsonStor

        async def fini():
            await self.jsonstor.fini()
            await slab.fini()

        self.onfini(fini)

    async def initServiceRuntime(self):
        self.addActiveCoro(self._clearInactiveSessions)

        if self.isactive:
            # bootstrap a CA for our aha:network
            netw = self.conf.get('aha:network')
            if netw is not None:

                await self.genCaCert(netw)

                host = self.conf.get('aha:name')
                if host is not None:
                    await self._genHostCert(host, signas=netw)

                user = self.conf.get('aha:admin')
                if user is not None:
                    await self._genUserCert(user, signas=netw)

    async def _clearInactiveSessions(self):

        async for svc in self.getAhaSvcs():
            if svc.get('svcinfo', {}).get('online') is None:
                continue
            current_sessions = {s_common.guid(iden) for iden in self.dmon.sessions.keys()}
            svcname = svc.get('svcname')
            network = svc.get('svcnetw')
            linkiden = svc.get('svcinfo').get('online')
            if linkiden not in current_sessions:
                logger.debug(f'AhaCell activecoro tearing down [{svcname}.{network}]')
                await self.setAhaSvcDown(svcname, linkiden, network=network)

        # Wait until we are cancelled or the cell is fini.
        await self.waitfini()

    async def getAhaSvcs(self, network=None):
        path = ('aha', 'services')
        if network is not None:
            path = path + (network,)

        async for path, item in self.jsonstor.getPathObjs(path):
            yield item

    def _nameAndNetwork(self, name, network):

        if network is None:
            svcfull = name
            try:
                svcname, svcnetw = name.split('.', 1)
            except ValueError:
                raise s_exc.BadArg(name=name, arg='name',
                                   mesg='Name must contain at least one "."') from None
        else:
            svcname = name
            svcnetw = network
            svcfull = f'{name}.{network}'

        return svcname, svcnetw, svcfull

    @s_nexus.Pusher.onPushAuto('aha:svc:add')
    async def addAhaSvc(self, name, info, network=None):

        svcname, svcnetw, svcfull = self._nameAndNetwork(name, network)

        full = ('aha', 'svcfull', svcfull)
        path = ('aha', 'services', svcnetw, svcname)

        unfo = info.get('urlinfo')
        logger.debug(f'Adding service [{svcfull}] from [{unfo.get("scheme")}://{unfo.get("host")}:{unfo.get("port")}]')

        svcinfo = {
            'name': svcfull,
            'svcname': svcname,
            'svcnetw': svcnetw,
            'svcinfo': info,
        }

        await self.jsonstor.setPathObj(path, svcinfo)
        await self.jsonstor.setPathLink(full, path)

        # mostly for testing...
        await self.fire('aha:svcadd', svcinfo=svcinfo)

    @s_nexus.Pusher.onPushAuto('aha:svc:del')
    async def delAhaSvc(self, name, network=None):

        svcname, svcnetw, svcfull = self._nameAndNetwork(name, network)

        full = ('aha', 'svcfull', svcfull)
        path = ('aha', 'services', svcnetw, svcname)

        await self.jsonstor.delPathObj(path)
        await self.jsonstor.delPathObj(full)

        # mostly for testing...
        await self.fire('aha:svcdel', svcname=svcname, svcnetw=svcnetw)

    async def setAhaSvcDown(self, name, linkiden, network=None):
        svcname, svcnetw, svcfull = self._nameAndNetwork(name, network)
        path = ('aha', 'services', svcnetw, svcname)

        svcinfo = await self.jsonstor.getPathObjProp(path, 'svcinfo')
        if svcinfo.get('online') is None:
            return

        await self._push('aha:svc:down', name, linkiden, network=network)

    @s_nexus.Pusher.onPush('aha:svc:down')
    async def _setAhaSvcDown(self, name, linkiden, network=None):
        svcname, svcnetw, svcfull = self._nameAndNetwork(name, network)
        path = ('aha', 'services', svcnetw, svcname)
        await self.jsonstor.cmpDelPathObjProp(path, 'svcinfo/online', linkiden)

        # Check if we have any links which may need to be removed
        current_sessions = {s_common.guid(iden): sess for iden, sess in self.dmon.sessions.items()}
        sess = current_sessions.get(linkiden)
        if sess is not None:
            for link in [lnk for lnk in self.dmon.links if lnk.get('sess') is sess]:
                await link.fini()

        await self.fire('aha:svcdown', svcname=svcname, svcnetw=svcnetw)

        logger.debug(f'Set [{svcfull}] offline.')

    async def getAhaSvc(self, name):
        path = ('aha', 'svcfull', name)
        svcinfo = await self.jsonstor.getPathObj(path)
        if svcinfo is not None:
            return svcinfo

    async def genCaCert(self, network):

        path = self.certdir.getCaCertPath(network)
        if path is not None:
            with open(path, 'rb') as fd:
                return fd.read().decode()

        logger.info(f'Generating CA certificate for {network}')
        fut = s_coro.executor(self.certdir.genCaCert, network, save=False)
        pkey, cert = await fut

        cakey = self.certdir._pkeyToByts(pkey).decode()
        cacert = self.certdir._certToByts(cert).decode()

        # nexusify storage..
        await self.saveCaCert(network, cakey, cacert)

        return cacert

    async def _genHostCert(self, hostname, signas=None):

        if os.path.isfile(os.path.join(self.dirn, 'certs', 'hosts', '{hostname}.crt')):
            return

        pkey, cert = await s_coro.executor(self.certdir.genHostCert, hostname, signas=signas, save=False)
        pkey = self.certdir._pkeyToByts(pkey).decode()
        cert = self.certdir._certToByts(cert).decode()
        await self.saveHostCert(hostname, pkey, cert)

    async def _genUserCert(self, username, signas=None):
        if os.path.isfile(os.path.join(self.dirn, 'certs', 'users', '{username}.crt')):
            return
        pkey, cert = await s_coro.executor(self.certdir.genUserCert, username, signas=signas, save=False)
        pkey = self.certdir._pkeyToByts(pkey).decode()
        cert = self.certdir._certToByts(cert).decode()
        await self.saveUserCert(username, pkey, cert)

    async def getCaCert(self, network):

        path = self.certdir.getCaCertPath(network)
        if path is None:
            return None

        with open(path, 'rb') as fd:
            return fd.read().decode()

    @s_nexus.Pusher.onPushAuto('aha:ca:save')
    async def saveCaCert(self, name, cakey, cacert):
        with s_common.genfile(self.dirn, 'certs', 'cas', f'{name}.key') as fd:
            fd.write(cakey.encode())
        with s_common.genfile(self.dirn, 'certs', 'cas', f'{name}.crt') as fd:
            fd.write(cacert.encode())

    @s_nexus.Pusher.onPushAuto('aha:host:save')
    async def saveHostCert(self, name, hostkey, hostcert):
        with s_common.genfile(self.dirn, 'certs', 'hosts', f'{name}.key') as fd:
            fd.write(hostkey.encode())

        with s_common.genfile(self.dirn, 'certs', 'hosts', f'{name}.crt') as fd:
            fd.write(hostcert.encode())

    @s_nexus.Pusher.onPushAuto('aha:user:save')
    async def saveUserCert(self, name, userkey, usercert):
        with s_common.genfile(self.dirn, 'certs', 'users', f'{name}.key') as fd:
            fd.write(userkey.encode())

        with s_common.genfile(self.dirn, 'certs', 'users', f'{name}.crt') as fd:
            fd.write(usercert.encode())

    async def signHostCsr(self, csrtext, signas=None, sans=None):
        xcsr = self.certdir._loadCsrByts(csrtext.encode())

        hostname = xcsr.get_subject().CN

        hostpath = s_common.genpath(self.dirn, 'certs', 'hosts', f'{hostname}.crt')
        if os.path.isfile(hostpath):
            os.unlink(hostpath)

        if signas is None:
            signas = hostname.split('.', 1)[1]

        logger.info(f'Signing host CSR for [{hostname}], signas={signas}, sans={sans}')

        pkey, cert = self.certdir.signHostCsr(xcsr, signas=signas, sans=sans)

        return self.certdir._certToByts(cert).decode()

    async def signUserCsr(self, csrtext, signas=None):
        xcsr = self.certdir._loadCsrByts(csrtext.encode())

        username = xcsr.get_subject().CN

        userpath = s_common.genpath(self.dirn, 'certs', 'users', f'{username}.crt')
        if os.path.isfile(userpath):
            os.unlink(userpath)

        if signas is None:
            signas = username.split('@', 1)[1]

        logger.info(f'Signing user CSR for [{username}], signas={signas}')

        pkey, cert = self.certdir.signUserCsr(xcsr, signas=signas)

        return self.certdir._certToByts(cert).decode()



import synapse.tests.utils as s_t_utils

import synapse.lib.output as s_output

class TestOutPut(s_t_utils.SynTest):

    # def test_output(self):
    #     outp = s_output.OutPut()
    #     outp.printf('foo')
    #     outp.printf('bar')

    def test_output_bytes(self):
        outp = s_output.OutPutBytes()

        outp.printf('foo')
        outp.printf('bar')

        outp.fd.seek(0)
        self.eq(outp.fd.read(), b'foo\nbar\n')

    def test_output_str(self):
        outp = s_output.OutPutStr()
        outp.printf('foo')
        outp.printf('bar')

        self.eq(str(outp), 'foo\nbar\n')



import synapse.tests.utils as s_t_utils
import synapse.tools.cryo.list as s_cryolist

class CryoListTest(s_t_utils.SynTest):

    async def test_cryolist(self):

        async with self.getTestCryo() as cryo:

            items = [(None, {'key': i}) for i in range(20)]

            tank = await cryo.init('hehe')
            await tank.puts(items)

            cryourl = cryo.getLocalUrl()

            argv = [cryourl]
            retn, outp = await self.execToolMain(s_cryolist.main, argv)

            self.eq(0, retn)
            outp.expect(cryourl)
            outp.expect('hehe: ')
            outp.expect("'indx': 20,")
            outp.expect("'entries': 20}")



import io
import binascii

import synapse.tests.utils as s_t_utils

import synapse.lib.hashset as s_hashset

asdf = b'asdfasdf'
asdfhash = '6a204bd89f3c8348afd5c77c717a097a'
asdfhash_iden = '1c753abfe85b4cbe46584fa5b1834fa4'
asdfhashb = binascii.unhexlify(asdfhash)

class HashsetTest(s_t_utils.SynTest):
    def hashset_assertions(self, hset):
        '''
        Test assertions for the hashset

        Args:
            hset (s_hashset.HashSet):
        '''
        self.eq(hset.size, 8)

        guid, props = hset.guid()
        self.eq(guid, asdfhash_iden)
        self.isinstance(props, dict)
        self.eq(props.get('md5'), asdfhash)
        self.isin('sha1', props)
        self.isin('sha256', props)
        self.isin('sha512', props)
        self.eq(props.get('size'), 8)

        digests = hset.digests()
        self.len(4, digests)
        hdict = dict(digests)
        self.eq(hdict.get('md5'), asdfhashb)
        self.isin('sha1', hdict)
        self.isin('sha256', hdict)
        self.isin('sha512', hdict)

    def test_lib_hashset_base(self):
        hset = s_hashset.HashSet()
        hset.update(asdf)
        self.hashset_assertions(hset)

    def test_lib_hashset_eatfd(self):
        fd = io.BytesIO(asdf)
        hset = s_hashset.HashSet()
        hset.eatfd(fd)
        self.hashset_assertions(hset)



import os

import synapse.common as s_common

class CoreModule:

    confdefs = ()
    mod_name = None

    def __init__(self, core, conf=None):

        self.core = core
        self.model = core.model
        if self.mod_name is None:
            self.mod_name = self.getModName()

        # Avoid getModPath / getConfPath during __init__ since these APIs
        # will create directories. We do not need that behavior by default.
        self._modpath = os.path.join(self.core.dirn,
                                     'mods',
                                     self.getModName())
        self._confpath = os.path.join(self._modpath, 'conf.yaml')

        if conf is None:
            conf = {}

        if os.path.isfile(self._confpath):
            conf = s_common.yamlload(self._confpath)

        self.conf = s_common.config(conf, self.confdefs)

    def getStormCmds(self):  # pragma: no cover
        '''
        Module implementers may override this to provide a list of Storm
        commands which will be loaded into the Cortex.

        Returns:
            list: A list of Storm Command classes (not instances).
        '''
        return ()

    def getModelDefs(self):
        return ()

    def getConfPath(self):
        '''
        Get the path to the module specific config file (conf.yaml).

        Notes:
            This creates the parent directory for the conf.yaml file if it does
            not exist. This API exists to allow a implementor to get the conf
            path during initCoreModule and drop a example config if needed.
            One use case of that is for missing configuration values, an
            example config can be written to the file and a exception raised.

        Returns:
            str: Path to where the conf file is located at.
        '''
        self.getModDir()
        return self._confpath

    def getModDir(self):
        '''
        Get the path to the module specific directory.

        Notes:
            This creates the directory if it did not previously exist.

        Returns:
            str: The filepath to the module specific directory.
        '''
        return s_common.gendir(self._modpath)

    def getModName(self):
        '''
        Return the lowercased name of this module.

        Notes:
            This pulls the ``mod_name`` attribute on the class. This allows
            an implementer to set a arbitrary name for the module.  If this
            attribute is not set, it defaults to
            ``self.__class__.__name__.lower()`` and sets ``mod_name`` to
            that value.

        Returns:
            (str): The module name.
        '''
        ret = self.mod_name
        if ret is None:
            ret = self.__class__.__name__
        return ret.lower()

    def getModPath(self, *paths):
        '''
        Construct a path relative to this module's working directory.

        Args:
            *paths: A list of path strings

        Notes:
            This creates the module specific directory if it does not exist.

        Returns:
            (str): The full path (or None if no cortex dir is configured).
        '''
        dirn = self.getModDir()
        return s_common.genpath(dirn, *paths)

    async def preCoreModule(self):
        '''
        Module implementers may override this method to execute code
        immediately after a module has been loaded.

        Notes:
            The ``initCoreModule`` function is preferred for overriding
            instead of ``preCoreModule()``.

            No Cortex layer/storage operations will function in preCoreModule.

            Any exception raised within this method will halt additional
            loading of the module.

        Returns:
            None
        '''

    async def initCoreModule(self):
        '''
        Module implementers may override this method to initialize the
        module after the Cortex has completed and is accessible to perform
        storage operations.

        Notes:
            This is the preferred function to override for implementing custom
            code that needs to be executed during Cortex startup.

            Any exception raised within this method will remove the module from
            the list of currently loaded modules.

            This is called for modules after getModelDefs() and getStormCmds()
            has been called, in order to allow for model loading and storm
            command loading prior to code execution offered by initCoreModule.

            A failure during initCoreModule will not unload data model or storm
            commands registered by the module.

        Returns:
            None
        '''



import os
import copy
import logging
import argparse
import collections.abc as c_abc

import yaml
import fastjsonschema

import synapse.exc as s_exc

import synapse.lib.hashitem as s_hashitem

from fastjsonschema.exceptions import JsonSchemaValueException

logger = logging.getLogger(__name__)

re_iden = '^[0-9a-f]{32}$'

# Cache of validator functions
_JsValidators = {}  # type: ignore

def getJsSchema(confbase, confdefs):
    '''
    Generate a Synapse JSON Schema for a Cell using a pair of confbase and confdef values.

    Args:
        confbase (dict): A JSON Schema dictionary of properties for the object. This content has
                         precedence over the confdefs argument.
        confdefs (dict): A JSON Schema dictionary of properties for the object.

    Notes:
        This generated a JSON Schema draft 7 schema for a single object, which does not allow for
        additional properties to be set on it.  The data in confdefs is implementer controlled and
        is welcome to specify

    Returns:
        dict: A complete JSON schema.
    '''
    props = {}
    schema = {
        '$schema': 'http://json-schema.org/draft-07/schema#',
        'additionalProperties': False,
        'properties': props,
        'type': 'object'
    }
    props.update(confdefs)
    props.update(confbase)
    return schema

def getJsValidator(schema, use_default=True):
    '''
    Get a fastjsonschema callable.

    Args:
        schema (dict): A JSON Schema object.
        use_default (bool): Whether to insert "default" key arguments into the validated data structure.

    Returns:
        callable: A callable function that can be used to validate data against the json schema.
    '''
    if schema.get('$schema') is None:
        schema['$schema'] = 'http://json-schema.org/draft-07/schema#'

    # It is faster to hash and cache the functions here than it is to
    # generate new functions each time we have the same schema.
    key = s_hashitem.hashitem((schema, use_default))
    func = _JsValidators.get(key)
    if func:
        return func

    func = fastjsonschema.compile(schema, use_default=use_default)

    def wrap(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except JsonSchemaValueException as e:
            raise s_exc.SchemaViolation(mesg=e.message, name=e.name) from e

    _JsValidators[key] = wrap
    return wrap

jsonschematype2argparse = {
    'integer': int,
    'string': str,
    'boolean': bool,
    'number': float,
}

def make_envar_name(key, prefix=None):
    '''
    Convert a colon delimited string into an uppercase, underscore delimited string.

    Args:
        key (str): Config key to convert.
        prefix (str): Optional string prefix to prepend the the config key.

    Returns:
        str: The string to lookup against a envar.
    '''
    nk = f'{key.replace(":", "_")}'
    if prefix:
        nk = f'{prefix}_{nk}'
    return nk.upper()

class Config(c_abc.MutableMapping):
    '''
    Synapse configuration helper based on JSON Schema.

    Args:
        schema (dict): The JSON Schema (draft v7) which to validate
                       configuration data against.
        conf (dict): Optional, a set of configuration data to preload.
        envar_prefix (str): Optional, a prefix used when collecting configuration
                            data from environment variables.

    Notes:
        This class implements the collections.abc.MutableMapping class, so it
        may be used where a dictionary would otherwise be used.

        The default values provided in the schema must be able to be recreated
        from the repr() of their Python value.

        Default values are not loaded into the configuration data until
        the ``reqConfValid()`` method is called.

    '''
    def __init__(self,
                 schema,
                 conf=None,
                 envar_prefix=None,
                 ):
        self.json_schema = schema
        if conf is None:
            conf = {}
        self.conf = conf
        self._argparse_conf_names = {}
        self._argparse_conf_parsed_names = {}
        self.envar_prefix = envar_prefix
        self.validator = getJsValidator(self.json_schema)

    @classmethod
    def getConfFromCell(cls, cell, conf=None, envar_prefix=None):
        '''
        Get a Config object from a Cell directly (either the ctor or the instance thereof).

        Returns:
            Config: A Config object.
        '''
        schema = getJsSchema(cell.confbase, cell.confdefs)
        return cls(schema, conf=conf, envar_prefix=envar_prefix)

    def getArgParseArgs(self):

        argdata = []

        for (name, conf) in self.json_schema.get('properties').items():

            if conf.get('hideconf'):
                continue

            if conf.get('hidecmdl'):
                continue

            typename = conf.get('type')
            # only allow single-typed values to have command line arguments
            if not isinstance(typename, str):
                continue

            atyp = jsonschematype2argparse.get(conf.get('type'))
            if atyp is None:
                continue

            akwargs = {
                'help': conf.get('description', ''),
                'action': 'store',
                'type': atyp,
            }

            if atyp is bool:

                default = conf.get('default')
                if default is None:
                    logger.debug(f'Boolean type is missing default information. Will not form argparse for [{name}]')
                    continue
                default = bool(default)
                akwargs['type'] = yaml.safe_load
                akwargs['choices'] = [True, False]
                akwargs['help'] = akwargs['help'] + f' This option defaults to {default}.'

            parsed_name = name.replace(':', '-')
            replace_name = name.replace(':', '_')
            self._argparse_conf_names[replace_name] = name
            self._argparse_conf_parsed_names[name] = parsed_name
            argname = '--' + parsed_name
            argdata.append((argname, akwargs))

        return argdata

    def getCmdlineMapping(self):
        if not self._argparse_conf_parsed_names:
            # Giv a shot at populating the data
            _ = self.getArgParseArgs()
        return {k: v for k, v in self._argparse_conf_parsed_names.items()}

    def setConfFromOpts(self, opts):
        '''
        Set the opts for a conf object from a namespace object.

        Args:
            opts (argparse.Namespace): A Namespace object made from parsing args with an ArgumentParser
            made with getArgumentParser.

        Returns:
            None: Returns None.
        '''
        opts_data = vars(opts)
        for k, v in opts_data.items():
            if v is None:
                continue
            nname = self._argparse_conf_names.get(k)
            if nname is None:
                continue
            self.setdefault(nname, v)

    # Envar support methods
    def setConfFromEnvs(self):
        '''
        Set configuration options from environment variables.

        Notes:
            Environment variables are resolved from configuration options after doing the following transform:

            - Replace ``:`` characters with ``_``.
            - Add a config provided prefix, if set.
            - Uppercase the string.
            - Resolve the environment variable
            - If the environment variable is set, set the config value to the results of ``yaml.yaml_safeload()``
              on the value.

            Configuration values which have the ``hideconf`` value set to True are not resolved from environment
            variables.

        Examples:

            For the configuration value ``auth:passwd``, the environment variable is resolved as ``AUTH_PASSWD``.
            With the prefix ``cortex``, the the environment variable is resolved as ``CORTEX_AUTH_PASSWD``.

        Returns:
            None: Returns None.
        '''
        name2envar = self.getEnvarMapping()
        for name, envar in name2envar.items():
            envv = os.getenv(envar)
            if envv is not None:
                envv = yaml.safe_load(envv)
                resp = self.setdefault(name, envv)
                if resp == envv:
                    logger.debug(f'Set config valu from envar: [{envar}]')

    def getEnvarMapping(self):
        '''
        Get a mapping of config values to envars.

        Configuration values which have the ``hideconf`` value set to True are not resolved from environment
        variables.
        '''
        ret = {}
        for name, conf in self.json_schema.get('properties', {}).items():
            if conf.get('hideconf'):
                continue

            envar = make_envar_name(name, prefix=self.envar_prefix)
            ret[name] = envar
        return ret

    # General methods
    def reqConfValid(self):
        '''
        Validate that the loaded configuration data is valid according to the schema.

        Notes:
            The validation set does set any default values which are not currently
            set for configuration options.

        Returns:
            None: This returns nothing.
        '''
        try:
            self.validator(self.conf)
        except s_exc.SchemaViolation as e:
            logger.exception('Configuration is invalid.')
            raise s_exc.BadConfValu(mesg=f'Invalid configuration found: [{str(e)}]') from None
        else:
            return

    def reqConfValu(self, key):
        '''
        Get a configuration value.  If that value is not present in the schema
        or is not set, then raise an exception.

        Args:
            key (str): The key to require.

        Returns:
            The requested value.
        '''
        # Ensure that the key is in self.json_schema
        if key not in self.json_schema.get('properties', {}):
            raise s_exc.BadArg(mesg='Required key is not present in the configuration schema.',
                               key=key)

        # Ensure that the key is present in self.conf
        if key not in self.conf:
            raise s_exc.NeedConfValu(mesg='Required key is not present in configuration data.',
                                     key=key)

        return self.conf.get(key)

    def asDict(self):
        '''
        Get a copy of configuration data.

        Returns:
            dict: A copy of the configuration data.
        '''
        return copy.deepcopy(self.conf)

    # be nice...
    def __repr__(self):
        info = [self.__class__.__module__ + '.' + self.__class__.__name__]
        info.append(f'at {hex(id(self))}')
        info.append(f'conf={self.conf}')
        return '<{}>'.format(' '.join(info))

    # ABC methods
    def __len__(self):
        return len(self.conf)

    def __iter__(self):
        return self.conf.__iter__()

    def __delitem__(self, key):
        return self.conf.__delitem__(key)

    def __setitem__(self, key, value):
        # This explicitly doesn't do any type validation.
        # The type validation is done on-demand, in order to
        # allow a user to incrementally construct the config
        # from different sources before turning around and
        # doing a validation pass which may fail.
        return self.conf.__setitem__(key, value)

    def __getitem__(self, item):
        return self.conf.__getitem__(item)



'''
A data model focused on material objects.
'''
import synapse.lib.module as s_module

class MatModule(s_module.CoreModule):

    def getModelDefs(self):
        modl = {
            'types': (
                ('mat:item', ('guid', {}), {'doc': 'A GUID assigned to a material object.'}),
                ('mat:spec', ('guid', {}), {'doc': 'A GUID assigned to a material specification.'}),
                ('mat:specimage', ('comp', {'fields': (('spec', 'mat:spec'), ('file', 'file:bytes'))}), {}),
                ('mat:itemimage', ('comp', {'fields': (('item', 'mat:item'), ('file', 'file:bytes'))}), {}),
                # TODO add base types for mass / volume
            ),

            'forms': (

                ('mat:item', {}, (

                    ('name', ('str', {'lower': True}), {'doc': 'The human readable name of the material item.'}),

                    ('spec', ('mat:spec', {}), {
                        'doc': 'The mat:spec of which this item is an instance.',
                    }),

                    ('place', ('geo:place', {}), {'doc': 'The most recent place the item is known to reside.'}),
                    ('latlong', ('geo:latlong', {}), {'doc': 'The last known lat/long location of the node.'}),

                    ('loc', ('loc', {}), {
                        'doc': 'The geo-political location string for the node.',
                    }),

                    # TODO add baseline things like dimensions / mass / etc?
                )),

                ('mat:spec', {}, (
                    ('name', ('str', {'lower': True}), {'doc': 'The human readable name of the material spec.'}),
                )),

                ('mat:itemimage', {}, (
                    ('item', ('mat:item', {}), {'doc': 'The item contained within the image file.', 'ro': True, }),
                    ('file', ('file:bytes', {}), {'doc': 'The file containing an image of the item.', 'ro': True, }),
                )),

                ('mat:specimage', {}, (
                    ('spec', ('mat:spec', {}), {'doc': 'The spec contained within the image file.', 'ro': True, }),
                    ('file', ('file:bytes', {}), {'doc': 'The file containing an image of the spec.', 'ro': True, }),
                )),
            ),
        }
        name = 'mat'
        return ((name, modl), )



import io
import logging
import msgpack
import msgpack.fallback as m_fallback

import synapse.exc as s_exc

logger = logging.getLogger(__name__)

# Single Packer object which is reused for performance
pakr = msgpack.Packer(use_bin_type=True, unicode_errors='surrogatepass')
if isinstance(pakr, m_fallback.Packer):  # pragma: no cover
    logger.warning('msgpack is using the pure python fallback implementation. This will impact performance negatively.')
    pakr = None

# synapse.lib.msgpack.un uses a hardcoded subset of these arguments for speed
unpacker_kwargs = {
    'raw': False,
    'use_list': False,
    'strict_map_key': False,
    'max_buffer_size': 2**32 - 1,
    'unicode_errors': 'surrogatepass'
}

def en(item):
    '''
    Use msgpack to serialize a compatible python object.

    Args:
        item (obj): The object to serialize

    Notes:
        String objects are encoded using utf8 encoding.  In order to handle
        potentially malformed input, ``unicode_errors='surrogatepass'`` is set
        to allow encoding bad input strings.

    Returns:
        bytes: The serialized bytes in msgpack format.
    '''
    try:
        return pakr.pack(item)
    except TypeError as e:
        pakr.reset()
        mesg = f'{e.args[0]}: {repr(item)[:20]}'
        raise s_exc.NotMsgpackSafe(mesg=mesg) from e
    except Exception as e:
        pakr.reset()
        mesg = f'Cannot serialize: {repr(e)}:  {repr(item)[:20]}'
        raise s_exc.NotMsgpackSafe(mesg=mesg) from e

def _fallback_en(item):
    '''
    Use msgpack to serialize a compatible python object.

    Args:
        item (obj): The object to serialize

    Notes:
        String objects are encoded using utf8 encoding.  In order to handle
        potentially malformed input, ``unicode_errors='surrogatepass'`` is set
        to allow encoding bad input strings.

    Returns:
        bytes: The serialized bytes in msgpack format.
    '''
    try:
        return msgpack.packb(item, use_bin_type=True, unicode_errors='surrogatepass')
    except TypeError as e:
        mesg = f'{e.args[0]}: {repr(item)[:20]}'
        raise s_exc.NotMsgpackSafe(mesg=mesg) from e
    except Exception as e:
        mesg = f'Cannot serialize: {repr(e)}:  {repr(item)[:20]}'
        raise s_exc.NotMsgpackSafe(mesg=mesg) from e

# Redefine the en() function if we're in fallback mode.
if pakr is None:  # pragma: no cover
    en = _fallback_en

def un(byts):
    '''
    Use msgpack to de-serialize a python object.

    Args:
        byts (bytes): The bytes to de-serialize

    Notes:
        String objects are decoded using utf8 encoding.  In order to handle
        potentially malformed input, ``unicode_errors='surrogatepass'`` is set
        to allow decoding bad input strings.

    Returns:
        obj: The de-serialized object
    '''
    # This uses a subset of unpacker_kwargs
    return msgpack.loads(byts, use_list=False, raw=False, strict_map_key=False, unicode_errors='surrogatepass')

def isok(item):
    '''
    Returns True if the item can be msgpacked (by testing packing).
    '''
    try:
        en(item)
        return True
    except Exception:
        return False

def iterfd(fd):
    '''
    Generator which unpacks a file object of msgpacked content.

    Args:
        fd: File object to consume data from.

    Notes:
        String objects are decoded using utf8 encoding.  In order to handle
        potentially malformed input, ``unicode_errors='surrogatepass'`` is set
        to allow decoding bad input strings.

    Yields:
        Objects from a msgpack stream.
    '''
    unpk = msgpack.Unpacker(fd, **unpacker_kwargs)
    for mesg in unpk:
        yield mesg

def iterfile(path, since=-1):
    '''
    Generator which yields msgpack objects from a file path.

    Args:
        path: File path to open and consume data from.

    Notes:
        String objects are decoded using utf8 encoding.  In order to handle
        potentially malformed input, ``unicode_errors='surrogatepass'`` is set
        to allow decoding bad input strings.

    Yields:
        Objects from a msgpack stream.
    '''
    with io.open(path, 'rb') as fd:

        unpk = msgpack.Unpacker(fd, **unpacker_kwargs)

        for i, mesg in enumerate(unpk):
            if i <= since:
                continue

            yield mesg

class Unpk:
    '''
    An extension of the msgpack streaming Unpacker which reports sizes.

    Notes:
        String objects are decoded using utf8 encoding.  In order to handle
        potentially malformed input, ``unicode_errors='surrogatepass'`` is set
        to allow decoding bad input strings.
    '''
    def __init__(self):
        self.size = 0
        self.unpk = msgpack.Unpacker(**unpacker_kwargs)

    def feed(self, byts):
        '''
        Feed bytes to the unpacker and return completed objects.

        Args:
            byts (bytes): Bytes to unpack.

        Notes:
            It is intended that this function is called multiple times with
            bytes from some sort of a stream, as it will unpack and return
            objects as they are available.

        Returns:
            list: List of tuples containing the item size and the unpacked item.
        '''
        self.unpk.feed(byts)

        retn = []

        while True:

            try:
                item = self.unpk.unpack()
                tell = self.unpk.tell()
                retn.append((tell - self.size, item))
                self.size = tell

            except msgpack.exceptions.OutOfData:
                break

        return retn

def loadfile(path):
    '''
    Load and upack the msgpack bytes from a file by path.

    Args:
        path (str): The file path to a message pack file.

    Raises:
        msgpack.exceptions.ExtraData: If the file contains multiple objects.

    Returns:
        (obj): The decoded python object.
    '''
    with io.open(path, 'rb') as fd:
        return un(fd.read())

def dumpfile(item, path):
    '''
    Dump an object to a file by path.

    Args:
        item (object): The object to serialize.
        path (str): The file path to save.

    Returns:
        None
    '''
    with io.open(path, 'wb') as fd:
        fd.write(en(item))

def deepcopy(item):
    '''
    Copy a msgpack serializable by packing then unpacking it.
    For complex primitives, this runs in about 1/3 the time of
    copy.deepcopy()
    '''
    return un(en(item))



import asyncio

import synapse.lib.boss as s_boss
import synapse.lib.task as s_task
import synapse.tests.utils as s_test

class FakeUser:
    def __init__(self, name):
        self.name = name

class TaskTest(s_test.SynTest):

    def test_loop_none(self):
        self.none(s_task.loop())

    async def test_loop_nn(self):
        self.nn(s_task.loop())

    async def test_task_module(self):

        async with await s_boss.Boss.anit() as boss:

            user = FakeUser('visi')

            synt = await boss.promote('test', user, info={'hehe': 'haha'})

            self.eq(s_task.user(), user)
            self.eq(s_task.current(), synt)
            self.eq(s_task.username(), 'visi')

    async def test_taskvars(self):
        s_task.varset('test', 'foo')
        self.eq(s_task.varget('test'), 'foo')

        self.none(s_task.varget('nope'))

        s_task.vardefault('test2', lambda: [1, 2, 3])
        test2 = s_task.varget('test2')
        self.eq(test2, [1, 2, 3])
        test2.append(4)
        self.eq([1, 2, 3, 4], s_task.varget('test2'))

        async def taskfunc():
            self.none(s_task.varget('test'))
            self.eq([1, 2, 3], s_task.varget('test2'))
            s_task.varset('test', 42)

        await asyncio.create_task(taskfunc())

        self.eq(s_task.varget('test'), 'foo')



import asyncio
import logging

import synapse.exc as s_exc
import synapse.telepath as s_telepath

import synapse.lib.base as s_base
import synapse.lib.stormtypes as s_stormtypes

logger = logging.getLogger(__name__)

stormcmds = (
    {
        'name': 'service.add',
        'descr': 'Add a storm service to the cortex.',
        'cmdargs': (
            ('name', {'help': 'The name of the service.'}),
            ('url', {'help': 'The telepath URL for the remote service.'}),
        ),
        'cmdconf': {},
        'storm': '''
            $sdef = $lib.service.add($cmdopts.name, $cmdopts.url)
            $lib.print("added {iden} ({name}): {url}", iden=$sdef.iden, name=$sdef.name, url=$sdef.url)
        ''',
    },
    {
        'name': 'service.del',
        'descr': 'Remove a storm service from the cortex.',
        'cmdargs': (
            ('iden', {'help': 'The service identifier or prefix.'}),
        ),
        'cmdconf': {},
        'storm': '''
            $svcs = ()

            for $sdef in $lib.service.list() {
                if $sdef.iden.startswith($cmdopts.iden) {
                    $svcs.append($sdef)
                }
            }

            $count = $svcs.length()

            if $( $count = 1 ) {
                $sdef = $svcs.index(0)
                $lib.service.del($sdef.iden)
                $lib.print("removed {iden} ({name}): {url}", iden=$sdef.iden, name=$sdef.name, url=$sdef.url)
            } elif $( $count = 0 ) {
                $lib.print("No service found by iden: {iden}", iden=$cmdopts.iden)
            } else {
                $lib.print('Multiple matches found for {iden}.  Aborting delete.', iden=$cmdopts.iden)
            }
        ''',
    },
    {
        'name': 'service.list',
        'descr': 'List the storm services configured in the cortex.',
        'cmdconf': {},
        'storm': '''
            $lib.print("")
            $lib.print("Storm service list (iden, ready, name, service name, service version, url):")
            $count = $(0)
            for $sdef in $lib.service.list() {
                $url = $sdef.url
                $iden = $sdef.iden
                $name = $sdef.name
                $ready = $sdef.ready
                $sname = $sdef.svcname
                if $sname {} else { $sname = 'Unknown' }
                $svers = $sdef.svcvers
                if $svers {
                    $svers = $lib.str.join('.', $svers)
                } else {
                    $svers = 'Unknown'
                }
                $mesg="    {iden} {ready} ({name}) ({sname} @ {svers}): {url}"
                $lib.print(mesg=$mesg, iden=$iden, ready=$ready, name=$name, sname=$sname, svers=$svers, url=$url)
                $count = $( $count + 1 )
            }
            $lib.print("")
            $lib.print("{count} services", count=$count)
        ''',
    }
)

class StormSvc:
    '''
    The StormSvc mixin class used to make a remote storm service with commands.
    '''

    _storm_svc_name = 'noname'
    _storm_svc_vers = (0, 0, 1)
    _storm_svc_evts = {}  # type: ignore
    _storm_svc_pkgs = ()  # type: ignore

    async def getStormSvcInfo(self):
        # Users must specify the service name
        assert self._storm_svc_name != 'noname'
        return {
            'name': self._storm_svc_name,
            'vers': self._storm_svc_vers,
            'evts': self._storm_svc_evts,
            'pkgs': await self.getStormSvcPkgs(),
        }

    async def getStormSvcPkgs(self):
        return self._storm_svc_pkgs

class StormSvcClient(s_base.Base):
    '''
    A StormService is a wrapper for a telepath proxy to a service
    accessible from the storm runtime.
    '''
    async def __anit__(self, core, sdef):

        await s_base.Base.__anit__(self)

        self.core = core
        self.sdef = sdef

        self.iden = sdef.get('iden')
        self.name = sdef.get('name')  # Local name for the cortex
        self.svcname = ''  # remote name from the service
        self.svcvers = ''  # remove version from the service

        # service info from the server...
        self.info = None

        url = self.sdef.get('url')

        self.ready = asyncio.Event()

        self.proxy = await s_telepath.Client.anit(url, onlink=self._onTeleLink)
        self.onfini(self.proxy.fini)

    async def _runSvcInit(self):
        # Set the latest reference for this object to the remote svcname
        self.core.svcsbysvcname.pop(self.svcname, None)
        self.svcname = self.info['name']
        self.svcvers = self.info['vers']
        self.core.svcsbysvcname[self.svcname] = self

        try:
            await self.core._delStormSvcPkgs(self.iden)

        except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
            raise

        except Exception:
            logger.exception(f'_delStormSvcPkgs failed for service {self.name} ({self.iden})')

        # Register new packages
        for pdef in self.info.get('pkgs', ()):

            try:
                # push the svciden in the package metadata for later reference.
                pdef['svciden'] = self.iden
                await self.core._normStormPkg(pdef)
                await self.core._addStormPkg(pdef)

            except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
                raise

            except Exception:
                name = pdef.get('name')
                logger.exception(f'addStormPkg ({name}) failed for service {self.name} ({self.iden})')

        # Set events and fire as needed
        evts = self.info.get('evts')
        try:
            if evts is not None:
                self.sdef = await self.core.setStormSvcEvents(self.iden, evts)

        except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
            raise

        except Exception:
            logger.exception(f'setStormSvcEvents failed for service {self.name} ({self.iden})')

        try:
            if self.core.isactive:
                await self.core._runStormSvcAdd(self.iden)

        except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
            raise

        except Exception:
            logger.exception(f'service.add storm hook failed for service {self.name} ({self.iden})')

    async def _onTeleLink(self, proxy):

        clss = proxy._getClasses()

        names = [c.rsplit('.', 1)[-1] for c in clss]

        if 'StormSvc' in names:
            self.info = await proxy.getStormSvcInfo()
            await self._runSvcInit()

        async def unready():
            self.ready.clear()
            await self.core.fire("stormsvc:client:unready", iden=self.iden)

        proxy.onfini(unready)

        self.ready.set()



import copy
import synapse.lib.autodoc as s_autodoc
import synapse.lib.stormtypes as s_stormtypes

import synapse.tests.utils as s_t_utils

class AutodocTest(s_t_utils.SynTest):

    def test_rst(self):

        page = s_autodoc.RstHelp()

        page.addHead('test', lvl=0, link='fakeLink')
        page.addLines('test', 'words\n go', 'here')
        page.addHead('burma', lvl=1)
        page.addLines('burma', 'shave')
        text = page.getRstText()
        expected = '''

fakeLink

####
test
####

test
words\n go
here

*****
burma
*****

burma
shave'''
        self.eq(text, expected)

    def test_helpers(self):

        # callsig
        rtype = {
            'args': (
                {'name': 'foo',
                 'type': 'str',
                 'desc': 'The foos!'},
                {'name': 'bar',
                 'type': ['str', 'int'],
                 'desc': 'The bar.',
                 'default': None},
                {'name': '**kwargs',
                 'type': 'any',
                 'desc': 'Extra foobars.',
                 },
            )
        }
        callsig = s_autodoc.genCallsig(rtype)
        self.eq(callsig, '(foo, bar=None, **kwargs)')
        self.eq(s_autodoc.genCallsig({}), '()')
        self.eq(s_autodoc.genCallsig({'args': ({'name': 'beep'},)}), '(beep)')

        # get arg lines
        self.eq(s_autodoc.getArgLines({}), [])
        lines = s_autodoc.getArgLines(rtype)
        self.eq(lines, ['\n', 'Args:', '    foo (str): The foos!', '\n',
                        '    bar: The bar. The input type may one one of the following: ``str``, ``int``.',
                        '\n', '    \\*\\*kwargs (any): Extra foobars.', '\n'])

        lines = s_autodoc.getArgLines({'args': [{'name': 'cmplx', 'type': {}, 'desc': 'unsupported'}]})
        self.eq(lines, ['\n', 'Args:',
                        '    cmplx: unsupported The input type is derived from the declarative type ``{}``.',
                        '\n'])

        with self.raises(AssertionError):
            s_autodoc.getArgLines({'args': [{'name': 'newp', 'type': 1234, 'desc': 'newp'}]})

        # return lines
        lines = s_autodoc.getReturnLines('str')
        self.eq(lines, ('', 'Returns:', '    The type is ``str``.'))
        lines = s_autodoc.getReturnLines(('str', 'int'))
        self.eq(lines, ('', 'Returns:', '    The type may be one of the following: ``str``, ``int``.'))

        lines = s_autodoc.getReturnLines({'returns': {'type': 'str', 'desc': 'The str!'}})
        self.eq(lines, ('', 'Returns:', '    The str! The return type is ``str``.'))

        lines = s_autodoc.getReturnLines({'returns': {'type': ['str', 'boolean']}})
        self.eq(lines, ('', 'Returns:', '    The return type may be one of the following: ``str``, ``boolean``.'))

        lines = s_autodoc.getReturnLines({'returns': {'type': 'str', 'name': 'Yields'}})
        self.eq(lines, ('', 'Yields:', '    The return type is ``str``.'))

        lines = s_autodoc.getReturnLines({'returns': {'type': {}}})
        self.eq(lines, ('', 'Returns:', '    The return type is derived from the declarative type ``{}``.'))

        with self.raises(AssertionError):
            s_autodoc.getReturnLines({'returns': {'type': 1234}})

        # prepare lines
        text = '''
        Hello

        Notes:
            Beep beep goes the docs

        Examples:
            Words!::

                example stuff
        '''
        lines = s_autodoc.prepareRstLines(text)
        self.eq(lines, ['Hello', '', 'Notes:', '    Beep beep goes the docs', '', 'Examples:',
                        '    Words!::', '', '        example stuff', ''])

        # Whole thing
        libtst = s_t_utils.LibTst

        locls = copy.deepcopy(libtst._storm_locals)
        # Strip _funcname out. This is normally done in the registry code.
        [obj.get('type', {}).pop('_funcname', None) for obj in locls]
        doc = {
            'desc': s_stormtypes.getDoc(libtst, "err"),
            'path': ('lib',) + libtst._storm_lib_path,
            'locals': locls,
        }
        page = s_autodoc.RstHelp()
        page.addHead('Test')
        page.addLines('I am a line.')
        s_autodoc.docStormTypes(page, (doc,), linkprefix='test')
        text = page.getRstText()
        expected = '''
####
Test
####

I am a line.


.. _test-lib-test:

********
lib.test
********

LibTst for testing!



.. _test-lib-test-beep:

beep(valu)
==========

Example storm func.

Notes:
    It beeps strings!


Args:
    valu (str): The value to beep.



Returns:
    The beeped string. The return type is ``str``.'''
        self.eq(text, expected)

        # Now as a library
        page = s_autodoc.RstHelp()
        page.addHead('Test')
        page.addLines('I am a line.')
        s_autodoc.docStormTypes(page, (doc,), linkprefix='test', islib=True)
        text = page.getRstText()
        expected = '''
####
Test
####

I am a line.


.. _test-lib-test:

*********
$lib.test
*********

LibTst for testing!



.. _test-lib-test-beep:

$lib.test.beep(valu)
====================

Example storm func.

Notes:
    It beeps strings!


Args:
    valu (str): The value to beep.



Returns:
    The beeped string. The return type is ``str``.'''
        self.eq(text, expected)



import synapse.tests.utils as s_t_utils

import synapse.lib.hashitem as s_hashitem

class HashItemTest(s_t_utils.SynTest):

    def test_lib_hashitem(self):
        x = {
            'foo': ['bar', 'baz'],
            'boing': ('gniob', 'boing'),

            'lol': 10,
            'hehe': 'haha',
            'gronk': {
                'hurr': 30,
                'durr': 40,
            },
        }

        y = {
            'boing': ('gniob', 'boing'),
            'foo': ['bar', 'baz'],

            'lol': 10,
            'gronk': {
                'durr': 40,
                'hurr': 30,
            },
            'hehe': 'haha',
        }

        self.eq(s_hashitem.hashitem(x), s_hashitem.hashitem(y))
